from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import attr
import numpy as np

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.flora import FloraConfig
from avalon.datagen.world_creation.constants import DARK_SHORE
from avalon.datagen.world_creation.constants import FRESH_WATER
from avalon.datagen.world_creation.constants import SWAMP
from avalon.datagen.world_creation.constants import UNCLIMBABLE_BIOME_ID
from avalon.datagen.world_creation.types import Biome
from avalon.datagen.world_creation.types import Color
from avalon.datagen.world_creation.utils import hex_to_rgb


# create the "biomes". Considers elevation and moisture (distance from water), looks that up in a table,
# and does some adjustments
@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BiomeConfig:
    # a mapping to the config for each scenery object
    flora_config: Dict[str, FloraConfig]
    # corresponds directly to attributes for the WorldEnvironment object in Godot
    # in order to discover the name of a setting, simply modify it, save it, and look at the resulting file
    godot_env_config: Dict[str, Any]
    # corresponds directly to attributes for the ProceduralSky object in Godot
    # in order to discover the name of a setting, simply modify it, save it, and look at the resulting file
    godot_sky_config: Dict[str, Any]
    # corresponds directly to attributes for the DirectionalLight object in Godot
    # in order to discover the name of a setting, simply modify it, save it, and look at the resulting file
    godot_sun_config: Dict[str, Any]
    # elevaations will be rescaled to this range
    min_elevation: float
    max_elevation: float
    # min and max dryness values allowed on the land.
    min_dryness: float
    max_dryness: float
    # moisture values greater than this will be set to min_dryness
    water_cutoff: float
    # the spatial scale of the noise pattern to be used to make the biome edges less regular
    noise_scale: float
    # how much to scale the noise data by when changing biome edges.
    # high values will make the biomes more random (and less strictly determined by height and moisture)
    noise_magnitude: float
    # where to cut off the slope. Is technically in units of "squared slope"
    low_slope_threshold: float  # max of low matrix
    high_slope_threshold: float  # max of middle matrix
    # just sets the colors based on the above mapping
    color_map: Dict[int, Tuple[float, float, float]]
    # mapping for biomes. See the doc where this is defined and passed in.
    biome_matrix: np.ndarray
    # list of all possible biomes. Must define one biome for each id used in the biome_matrix
    biomes: Tuple[Biome, ...]
    # controls the noise that is used to globally vary the density of scenery.
    # this noise texture crosses biome borders, so it can make consistent patterns across them
    # this is the usual spatial scale parameter for noise
    global_foliage_noise_scale: float = 0.01  # bigger = noisier, patterns bigger
    # this is the minimum for the noise. The max will always be 1.0
    # the noise is MULTIPLIED with the other weights
    # so the lower you set this, the less foliage there will be in general
    global_foliage_noise_min: float = 0.7  # closer to 1 = no effect from global noise
    # max meters away from the ocean to assign to the dark shore biome
    dark_shore_distance: float = 3.0
    # max meters of elevation above sea level to assign to the dark shore biome
    dark_shore_height: float = 1.0
    # this controls whether lakes will end up creating coasts or not (if False they will not)
    is_fresh_water_included_in_moisture: bool = False  # True = interior sandy beaches
    # controls whether normals are set per vertex (True) or per face (False)
    is_normal_per_vertex: float = False
    # controls whether colors are set per vertex (True) or per face (False)
    is_color_per_vertex: float = False
    # max elevation above sea level to consider "beach"
    max_shore_height: float = 8.0
    # maximum distance from the ocean to consider "beach"
    # I've set this pretty high right now just to help you see what is going on
    max_shore_distance: float = 20.0
    # any coastal points that have a higher slope than this cutoff will NOT be considered beach
    # and will thus retain their original biome (per the appropriate matrix)
    beach_slope_cutoff: float = 1.0  # same squared slope unit
    # helps to prevent the beach from reaching into other adjacent biomes along the shoreline
    # 1.0 means "dont reach into other biomes at all", while 0.0 means "go crazy and reach in as far
    # as is determined by the max shore distance and height params
    beach_shrink_fraction: float = 0.75
    # over how many meters do we blend the height from the non-beach terrain into the beach
    beach_fade_distance: float = 4.0
    # there must be at least this many meters of contiguous shore points of a sufficiently gentle slope
    # in order for it to be counted as a beach
    min_beach_length: float = 10.0
    # much like the beach shore, except for the swamp biome around bodies of fresh water
    # (elevation and distance maximums)
    swamp_elevation_max: float = 3.0
    swamp_distance_max: float = 10.0
    # extra noise parameters that can alter the shape of swamp biomes to make them not so regular in their appearance
    # normal scale (spatial) and magnitude (amount) params
    swamp_noise_scale: float = 0.01
    swamp_noise_magnitude: float = 1.0
    # controls whether scenery items are added at all
    is_scenery_added: bool = True
    # controls whether the perlin noise terrain shader is enabled. It looks nice, but is a bit slow on the oculus
    is_terrain_noise_shader_enabled: bool = True
    # how many neighbors to sample when mixing colors with adjacent neighbors. Set to 0 to disable
    color_sampling_neighbor_count: int = 10
    # controls the radius for what counts as a neighbor for color mixing
    color_sampling_std_dev: float = 3.0
    # how much noise there is in the coastal biome
    coast_color_noise: float = 0.0
    # how much noise there is in the rock biome
    rock_color_noise: float = 0.0
    # specifies which biome is the coastal one
    coastal_id: int = 3
    # specifies which biome is the rock one
    rock_biome_ids: Tuple[int, ...] = (0,)
    # can add independent noise to each scenery object, but this is slow
    is_independent_noise_per_scenery: bool = False
    # just a reference to the ExportConfig
    export_config: Optional[ExportConfig] = None
    # to generate levels more quickly, can set this to restrict the number of points for distance
    # calculations. Something like 2000 is plenty for good calculations without slowing things down
    # Something like 500 will be really fast, but with minor visual quality issues (noise)
    max_kd_points: Optional[int] = None
    # at this slope, the biome is guaranteed to be cliff
    force_cliff_square_slope: float = 100.0
    # a global modifier on how many scenery objects will be placed
    foliage_density_modifier: float = 1.0
    # whether to do the more complex beach and shore erosion calculations.
    is_beach_computed: bool = True


# there are 3 grids below, 1 for each level of steepness (from low to high)
# fmt: off
# 20 x 20 grid based on this: http://www-cs-students.stanford.edu/~amitp/game-programming/polygon-map-generation/
#    /\
#    |
#    |
#  elevation
#
#            dryness ------->
_raw_biome_id_mapping = np.array(
    [
        # low slope biomes
        np.flip([
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],  #
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
            [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],  #
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
            [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7],
            [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7],
            [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7],  #
            [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
        ], 0),
        # medium slope biomes
        np.flip([
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],  #
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
            [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
            [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],  #
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
            [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
            [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7],
            [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7],  #
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ], 0),
        # high slope biomes
        np.flip([
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  #
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  #
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  #
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ], 0),
    ]
)
# fmt: on


def generate_biome_config(
    rand: np.random.Generator,
    export_config: ExportConfig,
    max_height: float,
    foliage_density_modifier: float = 1.0,
):
    COASTAL_ID = 3
    ROCK_IDS = (0,)

    COLORS = {
        "white": "#FFFFFF",
        "light-green": "#9BC57A",
        "med-green": "#89B764",
        "dark-green": "#7CAA57",
        "dark-brown": "#B09B57",
        "light-brown": "#D0B771",
        "dark-grey": "#757069",
        "sand": "#E7E0A7",
        "swamp": "#A48A30",
        "bare": "#B6B9BC",
    }

    BIOMES = (
        Biome(2, "water", COLORS["sand"], color_jitter=0.0, correlated_color_jitter=0.0),  # ocean blue
        Biome(COASTAL_ID, "coastal", COLORS["sand"], color_jitter=0.01, correlated_color_jitter=0.02),  # light brown
        Biome(ROCK_IDS[0], "bare", COLORS["bare"], color_jitter=0.0, correlated_color_jitter=0.02),
        # Biome(ROCK_IDS[1], "scorched", COLORS["dark-grey"], color_jitter=0.0, correlated_color_jitter=0.025),
        Biome(5, "tropical rain forest", COLORS["light-green"], color_jitter=0.01, correlated_color_jitter=0.02),
        Biome(6, "tropical seasonal forest", COLORS["med-green"], color_jitter=0.01, correlated_color_jitter=0.02),
        Biome(7, "temperate rain forest", COLORS["med-green"], color_jitter=0.01, correlated_color_jitter=0.02),
        Biome(8, "temperate deciduous forest", COLORS["dark-green"], color_jitter=0.01, correlated_color_jitter=0.02),
        Biome(9, "grassland", COLORS["dark-brown"], color_jitter=0.01, correlated_color_jitter=0.02),
        Biome(10, "temperate desert", COLORS["dark-brown"], color_jitter=0.005, correlated_color_jitter=0.02),
        Biome(11, "subtropical desert", COLORS["light-brown"], color_jitter=0.005, correlated_color_jitter=0.02),
        # other stuff
        Biome(
            UNCLIMBABLE_BIOME_ID, "unclimbable", COLORS["dark-grey"], color_jitter=0.0, correlated_color_jitter=0.01
        ),
        Biome(DARK_SHORE, "dark shore", COLORS["sand"], color_jitter=0.0, correlated_color_jitter=0.01),
        Biome(FRESH_WATER, "fresh water", COLORS["swamp"], color_jitter=0.0, correlated_color_jitter=0.01),
        Biome(SWAMP, "swamp", COLORS["swamp"], color_jitter=0.0, correlated_color_jitter=0.01),
    )

    flora_config = {
        # scenery
        "acacia": FloraConfig(
            "res://scenery/trees/acacia.tscn",
            default_scale=0.6,
            is_noise_shader_enabled=False,
            height_offset=-1.5,
            collision_extents=np.array([0.507, 6, 0.533]),
        ),
        "fir": FloraConfig(
            "res://scenery/trees/fir.tscn",
            default_scale=0.6,
            is_noise_shader_enabled=False,
            height_offset=-2.0,
            collision_extents=np.array([1.0, 5.56902, 1.0]),
        ),
        "maple_red": FloraConfig(
            "res://scenery/trees/maple_red.tscn",
            default_scale=0.8,
            is_noise_shader_enabled=False,
            height_offset=-2.25,
            collision_extents=np.array([0.65, 6.0, 0.65]),
        ),
        "maple_orange": FloraConfig(
            "res://scenery/trees/maple_orange.tscn",
            default_scale=0.8,
            is_noise_shader_enabled=False,
            height_offset=-2.25,
            collision_extents=np.array([0.65, 6.0, 0.65]),
        ),
        "maple_yellow": FloraConfig(
            "res://scenery/trees/maple_yellow.tscn",
            default_scale=0.8,
            is_noise_shader_enabled=False,
            height_offset=-2.25,
            collision_extents=np.array([0.65, 6.0, 0.65]),
        ),
        "palm": FloraConfig(
            "res://scenery/trees/palm.tscn",
            default_scale=0.8,
            is_noise_shader_enabled=False,
            height_offset=-2.5,
            collision_extents=np.array([0.507, 5.75, 0.533]),
        ),
        "fruit_tree_normal": FloraConfig(
            "res://scenery/trees/fruit_tree_normal.tscn", default_scale=0.8, is_noise_shader_enabled=False
        ),
        "fruit_tree_tropical": FloraConfig(
            "res://scenery/trees/fruit_tree_tropical.tscn", default_scale=0.8, is_noise_shader_enabled=True
        ),
        "bush": FloraConfig("res://scenery/bush.tscn", default_scale=0.4, is_shadowed=False, height_offset=-0.25),
        "flower_blue": FloraConfig(
            "res://scenery/flower_blue.tscn", default_scale=0.7, is_shadowed=False, height_offset=-0.075
        ),
        "flower_pink": FloraConfig(
            "res://scenery/flower_pink.tscn", default_scale=0.7, is_shadowed=False, height_offset=-0.075
        ),
        "flower_yellow": FloraConfig(
            "res://scenery/flower_yellow.tscn", default_scale=0.7, is_shadowed=False, height_offset=-0.075
        ),
        "mushroom": FloraConfig(
            "res://scenery/mushroom.tscn", default_scale=0.5, is_shadowed=False, height_offset=-0.05
        ),
        # items
        "boulder": FloraConfig("res://items/stone/boulder.tscn", default_scale=0.5),
        "stone": FloraConfig("res://items/stone/stone.tscn", default_scale=0.5),
        "rock": FloraConfig("res://items/stone/rock.tscn", default_scale=1.1),
        "log": FloraConfig("res://items/wood/log.tscn", default_scale=1.0),
        "stick": FloraConfig("res://items/wood/stick.tscn", default_scale=1.0),
        # animals, predators
        "wolf": FloraConfig("res://entities/animals/predators/wolf_246f.tscn", default_scale=1.0),
        "jaguar": FloraConfig("res://entities/animals/predators/jaguar_348f.tscn", default_scale=1.0),
        "hippo": FloraConfig("res://entities/animals/predators/hippo_204f.tscn", default_scale=1.0),
        "bee": FloraConfig("res://entities/animals/predators/bee_218f.tscn", default_scale=1.0),
        "bear": FloraConfig("res://entities/animals/predators/bear_230f.tscn", default_scale=1.0),
        "alligator": FloraConfig("res://entities/animals/predators/alligator_270f.tscn", default_scale=1.0),
        "snake": FloraConfig("res://entities/animals/predators/snake_170f.tscn", default_scale=1.0),
        "eagle": FloraConfig("res://entities/animals/predators/eagle.tscn", default_scale=1.0),
        "hawk": FloraConfig("res://entities/animals/predators/hawk.tscn", default_scale=1.0),
        # animals, prey
        "deer": FloraConfig("res://entities/animals/prey/deer.tscn", default_scale=1.0),
        "frog": FloraConfig("res://entities/animals/prey/frog.tscn", default_scale=1.0),
        "rabbit": FloraConfig("res://entities/animals/prey/rabbit.tscn", default_scale=1.0),
        "pigeon": FloraConfig("res://entities/animals/prey/pigeon.tscn", default_scale=1.0),
        "squirrel": FloraConfig("res://entities/animals/prey/squirrel.tscn", default_scale=1.0),
        "turtle": FloraConfig("res://entities/animals/prey/turtle.tscn", default_scale=1.0),
        "mouse": FloraConfig("res://entities/animals/prey/mouse.tscn", default_scale=1.0),
        "crow": FloraConfig("res://entities/animals/prey/crow.tscn", default_scale=1.0),
    }

    # see the godot docs for each of these attributes.
    # they are effectively passed straight through to the generated .tscn files.
    sky_config = {
        "resource_local_to_scene": True,
        "sky_top_color": Color(0.705882, 0.87451, 1, 1),
        "sky_horizon_color": Color(1, 0.858824, 0.694118, 1),
        "sky_curve": 0.206137,
        "ground_bottom_color": Color(1, 0.858824, 0.694118, 1),
        "ground_horizon_color": Color(1, 0.858824, 0.694118, 1),
        "ground_curve": 0.112084,
        "sun_color": Color(0.886275, 0.576471, 0.545098, 1),
        "sun_latitude": 160.0,
        "sun_longitude": 180.0,
        "sun_angle_max": 10.0,
        "sun_curve": 0.186603,
        "sun_energy": 5.0,
    }

    # see the godot docs for each of these attributes.
    # they are effectively passed straight through to the generated .tscn files.
    env_config = {
        "resource_local_to_scene": True,
        "background_mode": 2,
        "ambient_light_color": Color(1, 1, 1, 1),
        "ambient_light_energy": 0.7,
        "ambient_light_sky_contribution": 0.0,
        "fog_enabled": export_config.is_fog_enabled,
        "fog_color": Color(0.847059, 0.901961, 0.960784, 1),
        "fog_sun_color": Color(0.992157, 0.960784, 0.847059, 1),
        "fog_sun_amount": 1.0,
        "fog_depth_begin": 20.0,
        "fog_depth_end": 4000.0,
        "fog_depth_curve": 0.404918,
        "fog_transmit_enabled": True,
        "fog_height_min": 100.0,
        "fog_height_max": -90.0,
        "fog_height_curve": 3.03143,
        "tonemap_mode": 2 if export_config.is_postprocessing_enabled else 0,
        "tonemap_white": 12.0,
        "ssao_enabled": export_config.is_ssao_enabled,
        "dof_blur_far_enabled": export_config.is_blur_enabled,
        "dof_blur_far_distance": 200.0,
        "dof_blur_far_transition": 10.0,
        "dof_blur_far_amount": 0.06,
        "dof_blur_far_quality": 2,
        "adjustment_enabled": export_config.is_postprocessing_enabled,
        "adjustment_brightness": 1.1,
        "adjustment_contrast": 1.1,
        "adjustment_saturation": 1.05,
    }

    # see the godot docs for each of these attributes.
    # they are effectively passed straight through to the generated .tscn files.
    sun_config = {
        "light_energy": 0.5,
        "light_color": Color(0.988235, 0.945098, 0.72549, 1),
        "light_indirect_energy": 0.0,
        "light_specular": 1.0,
        "shadow_enabled": export_config.is_sun_shadow_enabled,
        "shadow_color": Color(0.141176, 0.164706, 0.360784, 1),
        "directional_shadow_max_distance": 500.0,
    }

    is_scenery_added = True
    if export_config.scenery_mode == "none":
        is_scenery_added = False

    if export_config.is_legacy_biome_based_on_natural_height_set is not None:
        is_biome_based_on_natural_height = export_config.is_legacy_biome_based_on_natural_height_set
    else:
        is_biome_based_on_natural_height = rand.uniform() < 0.5
    if is_biome_based_on_natural_height:
        min_elevation = 0.0
        max_elevation = max_height / 40.0
        max_shore_height = 8.0
        max_shore_distance = 8.0
        is_beach_computed = True
    else:
        is_beach_computed = rand.uniform() < 0.5

        max_shore_height = rand.uniform(0.0, 8.0)
        max_shore_distance = rand.uniform(0.0, 8.0)

        # randomize the biomes--we're sometimes in the natural world, other times it's all one biome
        is_biome_height_dependent = rand.uniform() < 0.5
        if is_biome_height_dependent:
            max_height_for_biomes = rand.uniform(20.0, 60.0)
            max_elevation = min([max_height / max_height_for_biomes, 1.0])
            min_elevation = 0.0
        else:
            min_elevation = rand.uniform(0.0, 0.9)
            max_elevation = min_elevation + rand.uniform(0.0, 0.4)
            if max_elevation > 1.0:
                max_elevation = 1.0

    # randomize the scale of the noise. The first setting is more local variation, second is more global
    if rand.uniform() < 0.3:
        global_foliage_noise_scale = 0.005
    else:
        global_foliage_noise_scale = 0.05

    biome_config = BiomeConfig(
        flora_config=flora_config,
        godot_env_config=env_config,
        godot_sky_config=sky_config,
        godot_sun_config=sun_config,
        min_elevation=min_elevation,
        max_elevation=max_elevation,
        min_dryness=0.3,
        max_dryness=1.0,
        water_cutoff=0.9,
        noise_scale=0.05,
        noise_magnitude=4.0,
        is_color_per_vertex=False,
        is_normal_per_vertex=False,
        color_map={x.id: hex_to_rgb(x.color) for x in BIOMES},
        biome_matrix=_raw_biome_id_mapping,
        biomes=BIOMES,
        low_slope_threshold=0.5,
        high_slope_threshold=1.5,
        global_foliage_noise_scale=global_foliage_noise_scale,
        global_foliage_noise_min=0.0,
        is_fresh_water_included_in_moisture=False,
        dark_shore_distance=2.0,
        dark_shore_height=1.0,
        max_shore_distance=max_shore_distance,
        max_shore_height=max_shore_height,
        beach_slope_cutoff=0.5,
        min_beach_length=10.0,
        beach_shrink_fraction=0.0,
        beach_fade_distance=1.0,
        swamp_elevation_max=3.0,
        swamp_distance_max=3.0,
        swamp_noise_scale=0.01,
        swamp_noise_magnitude=1.0,
        color_sampling_neighbor_count=5,
        color_sampling_std_dev=2.0,
        coast_color_noise=0.0,
        rock_color_noise=0.0,
        coastal_id=COASTAL_ID,
        rock_biome_ids=ROCK_IDS,
        foliage_density_modifier=foliage_density_modifier,
        max_kd_points=None if export_config.is_border_calculation_detailed else 500,
        is_terrain_noise_shader_enabled=export_config.is_terrain_shader_enabled,
        is_scenery_added=is_scenery_added,
        is_beach_computed=is_beach_computed,
    )

    return biome_config



import json
from pathlib import Path
from typing import Dict
from typing import Optional
from typing import TypedDict
from typing import cast

import numpy as np
from loguru import logger

from avalon.contrib.testing_utils import RequestFixture
from avalon.contrib.testing_utils import fixture
from avalon.contrib.testing_utils import use
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.tests.params import CANONICAL_SEED
from avalon.datagen.world_creation.tests.params import INDOOR_WORLD_CATALOG
from avalon.datagen.world_creation.tests.params import INVALID_BUILDING_CATALOG
from avalon.datagen.world_creation.tests.params import OUTDOOR_WORLD_CATALOG
from avalon.datagen.world_creation.tests.params import VALID_BUILDING_CATALOG
from avalon.datagen.world_creation.worlds.world import build_building

EMPTY_FILE_CHECKSUM = "d41d8cd98f00b204e9800998ecf8427e"


@fixture
def seed_() -> int:
    return CANONICAL_SEED


@fixture(params=VALID_BUILDING_CATALOG.keys())
def building_catalog_id_(request: RequestFixture) -> str:
    return cast(str, request.param)


@fixture(params=VALID_BUILDING_CATALOG.items())
@use(seed_)
def building_(request: RequestFixture, seed: int) -> Building:
    building_catalog_id, building_config = request.param
    logger.info(f"Generating building {building_catalog_id}")
    rand = np.random.default_rng(seed)
    return build_building(building_config, building_catalog_id, rand)


@fixture(params=INVALID_BUILDING_CATALOG.keys())
def incompatible_building_catalog_id_(request: RequestFixture) -> str:
    return cast(str, request.param)


@fixture(params=INDOOR_WORLD_CATALOG.keys())
def indoor_world_catalog_id_(request: RequestFixture) -> str:
    return cast(str, request.param)


class ChecksumManifest(TypedDict):
    snapshot_commit: str
    checksums: Dict[str, str]


def get_current_reference_manifest(reference_name: str, data_path: Optional[Path] = None) -> ChecksumManifest:
    if data_path is None:
        data_path = Path(__file__).parent / "data"
    return cast(ChecksumManifest, json.load(open(data_path / f"{reference_name}_manifest.json", "r")))


@fixture(scope="session")
def buildings_manifest_() -> ChecksumManifest:
    return get_current_reference_manifest("buildings")


@fixture(scope="session")
def indoor_worlds_manifest_() -> ChecksumManifest:
    return get_current_reference_manifest("indoor_worlds")


@fixture(params=OUTDOOR_WORLD_CATALOG.keys())
def outdoor_world_catalog_id_(request: RequestFixture) -> str:
    return cast(str, request.param)


@fixture(scope="session")
def outdoor_worlds_manifest_() -> ChecksumManifest:
    return get_current_reference_manifest("outdoor_worlds")



import math
from enum import Enum
from typing import Set
from typing import Tuple
from typing import cast

import attr
import numpy as np
from scipy.spatial import Delaunay

from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.godot_base_types import Vector2
from avalon.datagen.world_creation.types import Point3DNP


class Axis(Enum):
    """Values are loweracase for easy use in getattr/setattr on objects with xyz properties"""

    X = "x"
    Y = "y"
    Z = "z"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BuildingTile:
    x: int
    z: int


def global_to_local_coords(global_coords: Point3DNP, offset: Point3DNP) -> Point3DNP:
    assert len(global_coords) == len(offset) == 3
    return cast(Point3DNP, global_coords - offset)


def local_to_global_coords(local_coords: Point3DNP, offset: Point3DNP) -> Point3DNP:
    assert len(local_coords) == len(offset) == 3
    return cast(Point3DNP, local_coords + offset)


def euclidean_distance(position_a: BuildingTile, position_b: BuildingTile) -> float:
    return math.sqrt(abs(position_a.x - position_b.x) ** 2 + abs(position_a.z - position_b.z) ** 2)


def midpoint(point_a: Point3DNP, point_b: Point3DNP) -> Point3DNP:
    return cast(Point3DNP, np.min([point_a, point_b], axis=0) + abs(point_a - point_b) / 2)


def get_triangulation_edges(triangulation: Delaunay) -> Set[Tuple[np.ndarray, np.ndarray]]:
    edges = set()
    point_indices = ((0, 1), (0, 2), (1, 2))
    for simplex in triangulation.simplices:
        for start_idx, end_idx in point_indices:
            point1 = cast(np.ndarray, simplex[start_idx])
            point2 = cast(np.ndarray, simplex[end_idx])
            edge = (point1, point2) if point1 < point2 else (point2, point1)
            edges.add(edge)
    return edges


def squares_overlap(centroid_a: Vector2, size_a: float, centroid_b: Vector2, size_b: float) -> bool:
    ax = FloatRange(centroid_a.x - size_a / 2, centroid_a.x + size_a / 2)
    ay = FloatRange(centroid_a.y - size_a / 2, centroid_a.y + size_a / 2)
    bx = FloatRange(centroid_b.x - size_b / 2, centroid_b.x + size_b / 2)
    by = FloatRange(centroid_b.y - size_b / 2, centroid_b.y + size_b / 2)
    return bool(ax.overlap(bx) and ay.overlap(by))



from enum import Enum


class HingeSide(Enum):
    RIGHT = "RIGHT"
    LEFT = "LEFT"


class MountSlot(Enum):
    TOP = "TOP"
    BOTTOM = "BOTTOM"


class LatchingMechanics(Enum):
    NO_LATCH = "NO_LATCH"
    LATCH_ONCE = "LATCH_ONCE"
    AUTO_LATCH = "AUTO_LATCH"



import pickle as pkl
from functools import partial
from pathlib import Path
from typing import Any
from typing import Dict

import numpy as np
import torch

from avalon.agent.common.types import Algorithm
from avalon.agent.common.util import postprocess_uint8_to_float
from avalon.contrib.utils import make_deterministic
from avalon.datagen.world_creation.tests.params import CANONICAL_SEED


def groom_observation_for_agent(
    observation: Dict[str, torch.Tensor], num_workers: int = 1, device: torch.device = torch.device("cpu")
) -> Dict[str, torch.Tensor]:
    with torch.no_grad():
        torch_obs = {
            k: torch.tensor(v).broadcast_to((num_workers, *v.shape)).to(device=device) for k, v in observation.items()
        }
        return postprocess_uint8_to_float(torch_obs)


def _dump(contents: Any, target_path: Path) -> None:
    with open(target_path, "wb") as target_file:
        pkl.dump(contents, target_file)


def _dump_dict_values(d: Dict[Any, Any], target_path: Path) -> None:
    # Using pickle.dump on numpy, torch and other objects can attach non-pertinent state to them which makes checking
    # deterministic behaviour hard. Use this function instead for such cases.
    for k, v in d.items():
        if isinstance(v, torch.Tensor):
            torch.save(v, target_path.parent / f"{target_path.name}_{k}.pt")
        elif isinstance(v, np.ndarray):
            np.save(str(target_path.parent / f"{target_path.name}_{k}"), v)
        else:
            pkl.dump(v, open(target_path.parent / f"{target_path.name}_{k}.pkl", "wb"))


def _load(source_path: Path) -> Any:
    with open(source_path, "rb") as source_file:
        return pkl.load(source_file)


def run_deterministic_forward_pass(
    agent: Algorithm, observation: Dict[str, torch.Tensor], seed: int = CANONICAL_SEED
) -> Dict[str, torch.Tensor]:
    assert agent.params.num_workers == 1, "Can't guarantee determinism when num_workers > 1"
    num_workers = 1

    ready_for_new_step = torch.ones((num_workers,), dtype=torch.bool)
    device = next(agent.parameters()).device
    torch_dones = torch.zeros((num_workers,), dtype=torch.bool, device=device)
    torch_obs = groom_observation_for_agent(observation, num_workers, device)
    make_deterministic(seed)
    agent.reset_state()
    agent.action_space.seed(seed)
    with torch.no_grad():
        action, _ = agent.rollout_step(torch_obs, torch_dones, ready_for_new_step, exploration_mode="eval")
    return action


assert_tensors_equal = partial(torch.testing.assert_close, rtol=0, atol=0)



from __future__ import (
    annotations,  # using this to get Postponed Evaluation of Annotations -- https://www.python.org/dev/peps/pep-0563/
)

import multiprocessing
from contextlib import contextmanager
from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import Any
from typing import ContextManager
from typing import Dict
from typing import Generator
from typing import List
from typing import Set
from typing import Type
from typing import TypeVar
from typing import Union
from uuid import UUID

import attr

FREEZE_KEY = "$_frozen"

TC = TypeVar("TC", bound="Serializable")


_CLASS_KEY = "$type"

lock = multiprocessing.Lock()


@attr.s(hash=True, collect_by_mro=True)
class Serializable:
    # noinspection PyDataclass
    def __attrs_post_init__(self) -> None:
        self.__dict__[FREEZE_KEY] = True

    def __setattr__(self, item: str, value: Any) -> None:
        if self.__dict__.get(FREEZE_KEY):
            raise ValueError("instance is frozen; see: .mutable_clone()")
        else:
            self.__dict__[item] = value

    def mutable_clone(self: TC) -> ContextManager[TC]:
        freeze_list: List[TC] = []

        def thaw(attr_object: TC) -> TC:
            kwargs: Dict[str, Any] = {}
            for attribute in attr.fields(attr_object.__class__):
                k = attribute.name
                v = getattr(attr_object, k)
                if isinstance(v, Serializable):
                    kwargs[k] = thaw(v)  # type: ignore
                elif isinstance(v, tuple) and len(v) > 0 and isinstance(v[0], Serializable):
                    kwargs[k] = tuple([thaw(x) for x in v])
                else:
                    kwargs[k] = deepcopy(v)

            # noinspection PyArgumentList
            attr_cloned = attr_object.__class__(**kwargs)
            attr_cloned.__dict__[FREEZE_KEY] = False
            freeze_list.append(attr_cloned)

            return attr_cloned

        @contextmanager
        def context() -> Generator[TC, None, None]:
            try:
                yield thaw(self)
            finally:
                for attr_cloned in freeze_list:
                    attr_cloned.__dict__[FREEZE_KEY] = True

        # noinspection PyTypeChecker
        return context()

    def to_dict(self) -> dict:
        dump: Dict[str, Any] = {_CLASS_KEY: get_qualname_from_serializable_type(type(self))}
        for attribute in attr.fields(self.__class__):
            k = attribute.name
            v = getattr(self, k)
            dump[k] = _to_dict(v)
        return dump

    @classmethod
    def from_dict(cls: Type[TC], dump: dict, is_upgrade_allowed: bool = False) -> TC:
        """
        Using is_upgrade_allowed is dangerous:
        - Silently ignores keys that exist in the serialization but no longer in the object
          Could definitely lose data.
        - Silently ignores missing attributes, assuming they will have default values.
          This will likely blow up if there are no default values.
        The most likely case that this will be problematic is if you RENAMED an attribute
        There's no way for us to really detect that.
        I'd be ok with someone adding explicit "rename" support when they actually need it
        """
        kwargs = {}
        remaining_dict_keys = set(dump.keys())
        for attribute in attr.fields(cls):
            k = attribute.name
            if is_upgrade_allowed and k not in dump:
                # this enables us to load older ubjects when new default attributes have been added
                # if there is no default value, the later construction will fail, there is
                # no easy way around that without getting into more complex migration schemes
                continue
            if k not in remaining_dict_keys:
                raise Exception(f"Serialized object {cls} is missing key={k}")
            remaining_dict_keys.remove(k)
            kwargs[k] = _from_value(dump[k], is_upgrade_allowed)
        if _CLASS_KEY in remaining_dict_keys:
            remaining_dict_keys.remove(_CLASS_KEY)
        if len(remaining_dict_keys) > 0:
            # silently drops keys here!
            if not is_upgrade_allowed:
                bad_keys = sorted(remaining_dict_keys)
                raise ParamTypeError(f"Tried to load data for {cls} but got unexpected keys: {bad_keys}")
        # noinspection PyArgumentList
        return cls(**kwargs)


def _to_dict(v: Any) -> Any:
    # better to check for the existance of this attribute than to check isinstance
    # even works in jupyter notebooks when doing code reloading then..
    if hasattr(v, "to_dict"):
        result = v.to_dict()
    elif isinstance(v, tuple):
        if len(v) > 0:
            result = tuple([_to_dict(x) for x in v])
        else:
            result = v
    elif isinstance(v, Enum):
        result = {"value": v.value, _CLASS_KEY: get_qualname_from_serializable_type(type(v))}
    elif isinstance(v, UUID):
        result = dict(value=str(v))
        result[_CLASS_KEY] = UUID.__name__
    elif isinstance(v, datetime):
        result = dict(value=v.isoformat())
        result[_CLASS_KEY] = datetime.__name__
    else:
        assert isinstance(v, (float, int, bool, str, type(None))), "Unexpected type: " + str(v)
        result = v
    return result


def _from_value(v: Dict[str, Any], is_upgrade_allowed: bool) -> Any:
    result: Any
    if isinstance(v, Dict) and _CLASS_KEY in v:
        class_name = v[_CLASS_KEY]
        if class_name == UUID.__name__:
            result = UUID(v["value"])
        elif class_name == datetime.__name__:
            result = datetime.fromisoformat(v["value"])
        else:
            t = get_serializable_type_from_qualname(class_name)
            if issubclass(t, Enum):
                # noinspection PyArgumentList
                result = t(value=v["value"])  # type: ignore
            else:
                result = t.from_dict(v, is_upgrade_allowed=is_upgrade_allowed)
    # TAKS c00f9a38-bd11-4e14-97d8-02b9bf843ca6: delete this after we get rid of OldSerializable
    elif isinstance(v, Dict) and "_serializable_type" in v:
        t = get_serializable_type_from_qualname(v["_serializable_type"])
        result = t.from_dict(v, is_upgrade_allowed=is_upgrade_allowed)
    elif isinstance(v, (list, tuple)):
        if len(v) == 0:
            result = tuple()
        else:
            inner_objects = []
            for inner_value in v:
                inner_objects.append(_from_value(inner_value, is_upgrade_allowed))
            result = tuple(inner_objects)
    else:
        assert isinstance(v, (float, int, bool, str, type(None))), "Unexpected type: " + str(v)
        result = v
    return result


qualname_to_serializable_type: Dict[str, Type[Serializable]] = {}
serializable_type_to_qualname: Dict[Type[Serializable], str] = {}
# this is here because some libraries make enums with the same names, but we dont use those, so, fine
unserializable_types: Set[str] = set()


class ParamTypeError(TypeError):
    pass


def _get_all_subclasses(cls: Any) -> List[Type[Serializable]]:
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(_get_all_subclasses(subclass))
    return all_subclasses


def _compute_unserializable(qualnames: List[str]) -> None:
    global unserializable_types

    if not unserializable_types:
        already_seen_names = set()
        for qualname in qualnames:
            if qualname in already_seen_names:
                unserializable_types.add(qualname)
            already_seen_names.add(qualname)


def _dedupe_subclasses_by_qualname(subclasses: List[type]) -> List[type]:
    qualname_to_serializable_type = {x.__qualname__: x for x in subclasses}
    return list(qualname_to_serializable_type.values())


def _dedupe_subclasses_by_id(subclasses: List[type]) -> List[type]:
    id_to_serializable_type = {id(x): x for x in subclasses}
    return list(id_to_serializable_type.values())


def get_all_serializable_classes() -> List[type]:
    # we don't reload the Enum class (since it's standard library) so we end up having duplicate subclasses for Enums
    enum_subclasses = _dedupe_subclasses_by_qualname(_get_all_subclasses(Enum))
    # we want to make sure we have classes with different names which is why we need to dedupe with the ids
    serializable_subclasses = _dedupe_subclasses_by_id(_get_all_subclasses(Serializable))
    return serializable_subclasses + enum_subclasses


def get_serializable_type_from_qualname(qualname: str) -> Type[Serializable]:
    global qualname_to_serializable_type
    global unserializable_types
    if not qualname_to_serializable_type:
        # makes our globals thread safe and hopefully free from weird race conditions
        with lock:
            if not qualname_to_serializable_type:
                all_serializable_classes = get_all_serializable_classes()
                qualname_to_serializable_type = {x.__qualname__: x for x in all_serializable_classes}
                _compute_unserializable([x.__qualname__ for x in all_serializable_classes])
    if qualname in unserializable_types:
        raise Exception(f"{qualname} cannot be deserialized because there are multiple definitions")
    return qualname_to_serializable_type[qualname]


def get_qualname_from_serializable_type(serializable_type: type) -> str:
    global serializable_type_to_qualname
    global unserializable_types
    if not serializable_type_to_qualname:
        # makes our globals thread safe and hopefully free from weird race conditions
        with lock:
            if not serializable_type_to_qualname:
                all_serializable_classes = get_all_serializable_classes()
                serializable_type_to_qualname = {x: x.__qualname__ for x in all_serializable_classes}
                _compute_unserializable(list(serializable_type_to_qualname.values()))
    result = serializable_type_to_qualname.get(serializable_type, serializable_type.__qualname__)
    if result in unserializable_types:
        raise Exception(f"{result} cannot be serialized because there are multiple definitions")
    return result


T = TypeVar("T", bool, int, float, str, tuple)
DictNest = Dict[str, Union[T, Dict[str, Any]]]
DictFlat = Dict[str, T]


def flatten_dict(d: DictNest, prefix: str = "") -> DictFlat:
    flattened_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            flattened_dict.update(flatten_dict(v, f"{prefix}{k}."))
        else:
            flattened_dict[f"{prefix}{k}"] = v
    return flattened_dict


def inflate_dict(d: DictFlat) -> DictNest:
    inflated_dict: DictNest = {}
    for key, value in d.items():
        parts = key.split(".")
        d = inflated_dict
        for part in parts[:-1]:
            d = d.setdefault(part, {})
        d[parts[-1]] = value
    return inflated_dict



from pathlib import Path
from typing import Tuple

import attr
import numpy as np

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.animals import ALL_PREY_CLASSES
from avalon.datagen.world_creation.entities.food import NON_TREE_FOODS
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.tasks.eat import add_food_tree
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.compositional import CompositeTaskConfig
from avalon.datagen.world_creation.worlds.compositional import create_compositional_task
from avalon.datagen.world_creation.worlds.compositional import get_random_predator
from avalon.datagen.world_creation.worlds.compositional import remove_close_predators
from avalon.datagen.world_creation.worlds.difficulty import get_rock_probability
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.world import World


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class SurviveTaskConfig(CompositeTaskConfig):
    task: AvalonTask = AvalonTask.SURVIVE
    is_big_world_mode_probability: float = 0.2
    is_hunter_mode_probability_easy: float = 0.0
    is_hunter_mode_probability_hard: float = 0.5
    # how big to make worlds where we can only hunt
    # if much larger than this, we would have to spawn too many animals for the dnsity to be reasonable
    # for any human or agent to do in a short amount of time
    normal_max_hunting_world_length: float = 350.0
    big_max_hunting_world_length: float = 600.0
    # for the condition where there is fruit and prey
    normal_max_hunting_and_gathering_world_length: float = 350.0
    big_max_hunting_and_gathering_world_length: float = 600.0
    # how much food to create in hunting and gathering mode at difficulty 0.0 and 1.0 respectively
    gathering_food_density_easy: float = 0.0003
    gathering_food_density_hard: float = 0.0001
    gathering_food_density_std_dev: float = 0.00005
    #
    hunting_predator_density_easy: float = 0.00001
    hunting_predator_density_hard: float = 0.0008
    hunting_predator_density_std_dev: float = 0.0001
    hunting_prey_density_easy: float = 0.0005
    hunting_prey_density_hard: float = 0.0001
    hunting_prey_density_std_dev: float = 0.0001
    hunting_tool_density_easy: float = 0.0005
    hunting_tool_density_hard: float = 0.0004
    hunting_tool_density_std_dev: float = 0.0001
    hunting_forage_density_easy: float = 0.0002
    hunting_forage_density_hard: float = 0.00005
    hunting_forage_density_std_dev: float = 0.00005
    #
    gathering_predator_dist_easy: float = 20.0
    gathering_predator_dist_hard: float = 5.0
    gathering_prey_dist_easy: float = 5.0
    gathering_prey_dist_hard: float = 10.0
    gathering_tool_dist_easy: float = 5.0
    gathering_tool_dist_hard: float = 10.0
    gathering_forage_food_dist_easy: float = 10.0
    gathering_forage_food_dist_hard: float = 30.0
    gathering_predator_count_distribution: Tuple[int, ...] = (0, 1, 2, 3, 4)
    gathering_prey_count_distribution: Tuple[int, ...] = (2, 1, 0, 0)
    gathering_tool_count_distribution: Tuple[int, ...] = (4, 2, 0)
    gathering_forage_food_count_distribution: Tuple[int, ...] = (2, 1, 0, 0)


def generate_survive_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: SurviveTaskConfig = SurviveTaskConfig(),
) -> None:
    is_hunter_mode_probability = scale_with_difficulty(
        difficulty, task_config.is_hunter_mode_probability_easy, task_config.is_hunter_mode_probability_hard
    )
    is_hunter_mode = rand.uniform() < is_hunter_mode_probability
    is_big_world_mode = rand.uniform() < task_config.is_big_world_mode_probability

    # how big to make worlds where we can only hunt
    # if much larger than this, we would have to spawn too many animals for the dnsity to be reasonable
    # for any human or agent to do in a short amount of time
    max_hunting_world_length = task_config.normal_max_hunting_world_length

    # for the condition where there is fruit and prey
    max_hunting_and_gathering_world_length = task_config.normal_max_hunting_and_gathering_world_length

    if is_big_world_mode:
        max_hunting_world_length = task_config.big_max_hunting_world_length
        max_hunting_and_gathering_world_length = task_config.big_max_hunting_and_gathering_world_length

    if is_hunter_mode:
        # if we are hunting, there are only animal foods. Harder because they cannot be seen so easily
        (
            output_difficulty,
            predator_density,
            prey_density,
            tool_density,
            forage_food_density,
        ) = get_open_world_densities(rand, difficulty, task_config)

        # we restrict the size of these worlds to keep the density high enough that you can actually hunt something...
        desired_goal_distance = scale_with_difficulty(difficulty, 20.0, max_hunting_world_length / 2.0)
        world, locations = create_compositional_task(
            rand,
            difficulty,
            task_config,
            export_config,
            desired_goal_distance,
            is_food_allowed=False,
            max_size_in_meters=max_hunting_world_length,
        )
        predator_count, prey_count, tool_count, forage_count = [
            get_count_from_density(world, x)
            for x in [predator_density, prey_density, tool_density, forage_food_density]
        ]

        output_difficulty, world = add_random_predators(rand, difficulty, world, predator_count)
        # noinspection PyUnusedLocal
        output_difficulty, world = add_random_prey(rand, difficulty, world, prey_count)
        world = add_random_tools(rand, world, tool_count, difficulty)
        world = add_forage_food(rand, world, forage_count)
    else:
        # if we're not hunters, they we can both hunt and gather
        # we'll put predators and prey in increasingly large radii around fruit trees
        world, locations = create_compositional_task(
            rand,
            difficulty,
            task_config,
            export_config,
            is_food_allowed=True,
            min_size_in_meters=scale_with_difficulty(difficulty, 20.0, max_hunting_and_gathering_world_length / 2.0),
            max_size_in_meters=max_hunting_and_gathering_world_length,
        )

        # this is number of food/predator categoricals generated per square meter
        food_density = normal_distrib_range(
            task_config.gathering_food_density_easy,
            task_config.gathering_food_density_hard,
            task_config.gathering_food_density_std_dev,
            rand,
            difficulty,
        )
        food_count = get_count_from_density(world, food_density)

        for i in range(food_count):
            difficulty, world = add_fruit_tree_and_animals(rand, difficulty, world, task_config)

    # since we added a bunch of predators, prevent them from being too close to our spanw
    world = remove_close_predators(world, to_2d_point(locations.spawn), task_config.min_predator_distance_from_spawn)

    export_world(output_path, rand, world)


def get_open_world_densities(
    rand: np.random.Generator, difficulty: float, task_config: SurviveTaskConfig
) -> Tuple[float, float, float, float, float]:
    # numbers calculated as how many of a thing in a 500 x 500 world (the largest)
    is_forage_food_available, difficulty = select_boolean_difficulty(difficulty, rand)

    # this is number of animals/tools per square meter
    predator_density = normal_distrib_range(
        task_config.hunting_predator_density_easy,
        task_config.hunting_predator_density_hard,
        task_config.hunting_predator_density_std_dev,
        rand,
        difficulty,
    )
    prey_density = normal_distrib_range(
        task_config.hunting_prey_density_easy,
        task_config.hunting_prey_density_hard,
        task_config.hunting_prey_density_std_dev,
        rand,
        difficulty,
    )
    tool_density = normal_distrib_range(
        task_config.hunting_tool_density_easy,
        task_config.hunting_tool_density_hard,
        task_config.hunting_tool_density_std_dev,
        rand,
        difficulty,
    )

    if is_forage_food_available:
        # this is number of food per square meter
        forage_food_density = normal_distrib_range(
            task_config.hunting_forage_density_easy,
            task_config.hunting_forage_density_hard,
            task_config.hunting_forage_density_std_dev,
            rand,
            difficulty,
        )
    else:
        forage_food_density = 0.0
    return difficulty, predator_density, prey_density, tool_density, forage_food_density


def get_count_from_density(world: World, density: float) -> int:
    square_meters = world.config.size_in_meters * world.config.size_in_meters
    count = round(square_meters * density) + 1
    return count


def add_random_predators(
    rand: np.random.Generator, difficulty: float, world: World, count: int
) -> Tuple[float, World]:
    for i in range(count):
        position = world.get_safe_point(rand, island_mask=None)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place predator")
        output_difficulty, world = create_predator(difficulty, position, rand, world)
    return difficulty, world


def create_predator(
    difficulty: float, position: Point3DNP, rand: np.random.Generator, world: World
) -> Tuple[float, World]:
    predator, difficulty = get_random_predator(rand, difficulty)
    predator = attr.evolve(predator, position=position)
    world = world.add_item(predator, reset_height_offset=predator.get_offset())
    return difficulty, world


def add_random_tools(rand: np.random.Generator, world: World, count: int, difficulty: float) -> World:
    for i in range(count):
        position = world.get_safe_point(rand, island_mask=None)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place tool")
        world = create_tool(position, rand, world, difficulty)
    return world


def add_forage_food(rand: np.random.Generator, world: World, count: int) -> World:
    for i in range(count):
        position = world.get_safe_point(rand, island_mask=None)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place food")
        world = create_forage_food(position, rand, world)
    return world


def create_tool(position: Point3DNP, rand: np.random.Generator, world: World, difficulty: float) -> World:
    tool: Tool
    if rand.uniform() < get_rock_probability(difficulty):
        tool = Rock()
    else:
        tool = Stick()
    tool = attr.evolve(tool, position=position)
    return world.add_item(tool, reset_height_offset=tool.get_offset())


def create_forage_food(position: Point3DNP, rand: np.random.Generator, world: World) -> World:
    food = rand.choice(NON_TREE_FOODS)  # type: ignore[arg-type]
    food = attr.evolve(food, position=position)
    return world.add_item(food, reset_height_offset=food.get_offset())


def add_random_prey(rand: np.random.Generator, difficulty: float, world: World, count: int) -> Tuple[float, World]:
    for i in range(count):
        position = world.get_safe_point(rand, island_mask=None)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place prey")
        output_difficulty, world = create_prey(difficulty, position, rand, world)
    return difficulty, world


def create_prey(
    difficulty: float, position: Point3DNP, rand: np.random.Generator, world: World
) -> Tuple[float, World]:
    prey, difficulty = get_random_prey(rand, difficulty)
    prey = attr.evolve(prey, position=position)
    world = world.add_item(prey, reset_height_offset=prey.get_offset())
    return difficulty, world


def add_fruit_tree_and_animals(
    rand: np.random.Generator, difficulty: float, world: World, task_config: SurviveTaskConfig
) -> Tuple[float, World]:
    tree_position = world.get_safe_point(rand, island_mask=None)
    if tree_position is None:
        raise ImpossibleWorldError("Couldn't find safe point to place fruit tree")
    output_difficulty, tool, world = add_food_tree(
        rand, difficulty, world, to_2d_point(tree_position), is_tool_food_allowed=True
    )
    sq_distances_from_tree = world.map.get_dist_sq_to(to_2d_point(tree_position))

    predator_count, output_difficulty = select_categorical_difficulty(
        task_config.gathering_predator_count_distribution, difficulty, rand
    )
    predator_dist = scale_with_difficulty(
        difficulty, task_config.gathering_predator_dist_easy, task_config.gathering_predator_dist_hard
    )
    safe_mask = world.get_safe_mask(
        sq_distances=sq_distances_from_tree, max_sq_dist=predator_dist**2, island_mask=None
    )
    for i in range(predator_count):
        position = world._get_safe_point(rand, safe_mask)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place predator")
        output_difficulty, world = create_predator(difficulty, position, rand, world)

    prey_count, output_difficulty = select_categorical_difficulty(
        task_config.gathering_prey_count_distribution, difficulty, rand
    )
    prey_dist = scale_with_difficulty(
        difficulty, task_config.gathering_prey_dist_easy, task_config.gathering_prey_dist_hard
    )
    safe_mask = world.get_safe_mask(sq_distances=sq_distances_from_tree, max_sq_dist=prey_dist**2, island_mask=None)
    for i in range(prey_count):
        position = world._get_safe_point(rand, safe_mask)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place prey")
        output_difficulty, world = create_prey(difficulty, position, rand, world)

    tool_count, output_difficulty = select_categorical_difficulty(
        task_config.gathering_tool_count_distribution, difficulty, rand
    )
    tool_dist = scale_with_difficulty(
        difficulty, task_config.gathering_tool_dist_easy, task_config.gathering_tool_dist_hard
    )
    safe_mask = world.get_safe_mask(sq_distances=sq_distances_from_tree, max_sq_dist=tool_dist**2, island_mask=None)
    for i in range(tool_count):
        position = world._get_safe_point(rand, safe_mask)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place tool")
        world = create_tool(position, rand, world, difficulty)

    forage_food_count, output_difficulty = select_categorical_difficulty(
        task_config.gathering_forage_food_count_distribution, difficulty, rand
    )
    forage_food_dist = scale_with_difficulty(
        difficulty, task_config.gathering_forage_food_dist_easy, task_config.gathering_forage_food_dist_hard
    )
    safe_mask = world.get_safe_mask(
        sq_distances=sq_distances_from_tree, max_sq_dist=forage_food_dist**2, island_mask=None
    )
    for i in range(forage_food_count):
        position = world._get_safe_point(rand, safe_mask)
        if position is None:
            raise ImpossibleWorldError("Couldn't find safe point to place food")
        world = create_forage_food(position, rand, world)

    return difficulty, world


def get_random_prey(rand: np.random.Generator, difficulty: float):
    prey_class, new_difficulty = select_categorical_difficulty(ALL_PREY_CLASSES, difficulty, rand)
    prey = prey_class(position=np.array([0.0, 0.0, 0.0]))
    return prey, new_difficulty



# mypy: ignore-errors
# TODO: type this file that's not used in production
from abc import abstractmethod
from typing import Callable

import gym
import numpy as np
import torch.nn as nn
from torch import Tensor
from torch.nn import Module

from avalon.agent.common.action_model import DictActionDist
from avalon.agent.common.action_model import DictActionHead
from avalon.agent.common.types import ObservationBatch


class PPOModel(Module):
    @abstractmethod
    def forward(self, obs: ObservationBatch) -> tuple[Tensor, DictActionDist]:
        raise NotImplementedError


def init(module: Module, weight_init: Callable, bias_init: Callable, gain: float = 1) -> Module:
    """Helper to initialize a layer weight and bias."""
    weight_init(module.weight.data, gain=gain)
    bias_init(module.bias.data)
    return module


def mlp_init(module: Module, gain: float = np.sqrt(2), bias: float = 0.0) -> Module:
    """Helper to initialize a layer weight and bias."""
    nn.init.orthogonal_(module.weight, gain=gain)
    nn.init.constant_(module.bias, bias)
    return module


class Flatten(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x.reshape(x.size(0), -1)


class CNNBase(PPOModel):
    def __init__(self, params, observation_space, action_space, hidden_size=512) -> None:
        """Initializer.
        num_channels: the number of channels in the input images (eg 3
            for RGB images, or 12 for a stack of 4 RGB images).
        num_outputs: the dimension of the output distribution.
        dist: the output distribution (eg Discrete or Normal).
        hidden_size: the size of the final actor+critic linear layers
        """
        super().__init__()

        # TODO: make this more general to allow non-scalar spaces
        assert len(observation_space["wrapped"].shape) == 3
        num_channels = observation_space["wrapped"].shape[0]
        self.num_outputs = len(action_space.spaces)

        init_ = lambda m: init(
            m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0), nn.init.calculate_gain("relu")
        )

        self.main = nn.Sequential(
            init_(nn.Conv2d(num_channels, 32, kernel_size=8, stride=4)),
            nn.ReLU(),
            init_(nn.Conv2d(32, 64, kernel_size=4, stride=2)),
            nn.ReLU(),
            init_(nn.Conv2d(64, 32, kernel_size=3, stride=1)),
            nn.ReLU(),
            Flatten(),
            # init_(nn.Linear(32 * 7 * 7, hidden_size)),
            # nn.ReLU(),
            mlp_init(nn.Linear(32 * 7 * 7, hidden_size)),
            nn.Tanh(),
        )

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))

        self.critic_linear = mlp_init(nn.Linear(hidden_size, 1), gain=1.0)
        self.action_head = DictActionHead(action_space, params)
        self.actor_linear = mlp_init(nn.Linear(hidden_size, self.action_head.num_inputs), gain=0.01)

    def forward(self, obs):
        """x should have shape (batch_size, num_channels, 84, 84)."""
        x = obs["wrapped"]
        batch_size = x.shape[0]
        x = self.main(x)
        value = self.critic_linear(x).reshape((batch_size,))
        action_logits = self.actor_linear(x)
        assert value.shape == (batch_size,)
        assert action_logits.shape == (batch_size, self.action_head.num_inputs)
        return value, self.action_head(action_logits)


class MLPBase(PPOModel):
    def __init__(self, params, observation_space, action_space, hidden_size=64) -> None:
        super().__init__()

        assert isinstance(observation_space, gym.spaces.Dict)
        assert len(observation_space) == 1
        self.key = list(observation_space.keys())[0]
        assert len(observation_space[self.key].shape) == 1
        num_inputs = observation_space[self.key].shape[0]

        self.action_head = DictActionHead(action_space, params)

        self.actor = nn.Sequential(
            mlp_init(nn.Linear(num_inputs, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, self.action_head.num_inputs), gain=0.01),
        )

        self.critic = nn.Sequential(
            mlp_init(nn.Linear(num_inputs, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, 1), gain=1.0),
        )

    def forward(self, obs):
        x = obs[self.key]
        batch_size = x.shape[0]
        value = self.critic(x).reshape((batch_size,))
        action_logits = self.actor(x)
        assert value.shape == (batch_size,)
        assert action_logits.shape == (batch_size, self.action_head.num_inputs)
        return value, self.action_head(action_logits)



from typing import List

import attr
import numpy as np

from avalon.datagen.world_creation.constants import IDENTITY_BASIS
from avalon.datagen.world_creation.entities.doors.locks.door_lock import DoorLock
from avalon.datagen.world_creation.entities.item import Item
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.types import Basis3DNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Door(Item):
    entity_id: int = -1

    # for doors, position is the door's centroid - the middle point of the plane we're trying to close using the door
    is_dynamic: bool = True
    size: np.ndarray = np.array([1, DEFAULT_STORY_HEIGHT, 0.1])
    rotation: Basis3DNP = IDENTITY_BASIS
    locks: List[DoorLock] = attr.field(default=attr.Factory(list))
    face_axis: Axis = Axis.X



from setuptools import setup
from setuptools.dist import Distribution


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""

    def has_ext_modules(self) -> bool:
        return True


setup(
    # Include pre-compiled extension
    # package_data={"packagename": ["_precompiled_extension.pyd"]},
    # distclass=BinaryDistribution,
)



import shutil
from collections import deque
from io import BufferedReader
from io import BufferedWriter
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Deque
from typing import Dict
from typing import Final
from typing import Generic
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import gym
import numpy as np
from loguru import logger
from numpy import typing as npt

from avalon.common.error_utils import capture_exception
from avalon.common.utils import only
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.datagen.errors import GodotError
from avalon.datagen.godot_env._bridge import GodotEnvBridge
from avalon.datagen.godot_env.action_log import GodotEnvActionLog
from avalon.datagen.godot_env.actions import ActionType
from avalon.datagen.godot_env.actions import DebugCameraAction
from avalon.datagen.godot_env.goals import GoalEvaluator
from avalon.datagen.godot_env.goals import GoalProgressResult
from avalon.datagen.godot_env.interactive_godot_process import InteractiveGodotProcess
from avalon.datagen.godot_env.observations import FeatureDataDict
from avalon.datagen.godot_env.observations import GodotObservationContext
from avalon.datagen.godot_env.observations import ObservationType
from avalon.datagen.godot_generated_types import SimSpec

# Mapping of feature name to (data_type, shape).
from avalon.datagen.world_creation.constants import STARTING_HIT_POINTS
from avalon.datagen.world_creation.world_generator import EmptyLevelGenerator
from avalon.datagen.world_creation.world_generator import GeneratedWorldParamsType
from avalon.datagen.world_creation.world_generator import WorldGenerator

_DEFAULT_EPISODE_SEED: Final = 0

if TYPE_CHECKING:
    from avalon.datagen.godot_env.replay import GodotEnvReplay


class GodotEnv(gym.Env, Generic[ObservationType, ActionType, GeneratedWorldParamsType]):
    """An OpenAI Gym Env that communicates with Godot over unix pipes.

    We expose the main gym.Env methods, although reset does not accept all arguments yet:
        step
        reset
        render
        close
        seed

    There are also two better-typed alternatives to step and render, act and reset_nicely,
    which accept and return the supplied ActionType and ObservationType.

    The python process must communicate in lock-step with godot, which can naturally be
    delicate and result in deadlock.
    """

    # the gym.Env#step interface returns an observation, which could
    # be something other than the screen state itself.
    #
    # To provide this affordance in the future, we could add observation shape to our SimSpec
    # and double tap the pipe on the godot side with ...observation,screen...
    # agent_observes_screen = True

    _action_pipe: BufferedWriter
    _screen_pipe: BufferedReader
    # TODO numpy array
    _current_screen: Optional[npt.NDArray]  # numpy.ndarray(x, y, 3)
    _screen_buffer_size: int
    _latest_screen: Optional[npt.NDArray]

    def __init__(
        self,
        config: SimSpec,
        observation_type: Type[ObservationType],
        action_type: Type[ActionType],
        goal_evaluator: GoalEvaluator[ObservationType, GeneratedWorldParamsType],
        gpu_id: int = 0,
        is_logging_artifacts_on_error_to_s3: bool = False,
        s3_bucket_name: Optional[str] = None,
        is_error_log_checked_after_each_step: bool = True,
        is_observation_space_flattened: bool = False,
        is_godot_restarted_on_error: bool = False,
        is_dev_flag_added: bool = False,
        run_uuid: Optional[str] = None,
    ) -> None:
        self.config = config
        self.action_type = action_type
        self.goal_evaluator = goal_evaluator

        self.action_space = self.action_type.to_gym_space()

        self.is_dev_flag_added = is_dev_flag_added
        self.is_logging_artifacts_on_error_to_s3 = is_logging_artifacts_on_error_to_s3
        if self.is_logging_artifacts_on_error_to_s3:
            assert s3_bucket_name is not None, "Must provide S3 bucket name"
        self.s3_bucket_name = s3_bucket_name
        self.is_godot_restarted_on_error = is_godot_restarted_on_error
        self.is_reset_called_already: bool = False  # :(
        self._latest_screen: Optional[npt.NDArray] = None
        self._latest_episode_seed: int = 0

        self.is_error_log_checked_after_each_step = is_error_log_checked_after_each_step

        assert isinstance(self.config, SimSpec), "cannot establish godot pipe without fixed resolution"

        # TODO: this code is getting pretty duplicated and bad...
        self.world_generator = self._create_world_generator()

        self.gpu_id = gpu_id
        self.process = InteractiveGodotProcess(
            self.config, is_dev_flag_added=is_dev_flag_added, gpu_id=self.gpu_id, run_uuid=run_uuid
        )
        self._bridge: GodotEnvBridge[ActionType] = GodotEnvBridge.build_by_starting_process(
            self.process,
            screen_resolution=(self.config.recording_options.resolution_x, self.config.recording_options.resolution_y),
        )

        self.observation_context = GodotObservationContext(
            observation_type=observation_type,
            is_space_flattened=is_observation_space_flattened,
            available_features=self._bridge.query_available_features(),
        )
        self._bridge.select_and_cache_features(self.observation_context.selected_features)
        self.seed_nicely(0)
        self.recent_worlds: Deque[GeneratedWorldParamsType] = deque()

    def _create_world_generator(self) -> WorldGenerator[GeneratedWorldParamsType]:
        return cast(
            WorldGenerator[GeneratedWorldParamsType], EmptyLevelGenerator(base_path=Path("/tmp/level_gen"), seed=0)
        )

    def _restart_godot_quietly(self) -> None:
        if self.is_running:
            if self._bridge.is_open:
                try:
                    self._bridge.close()
                except BrokenPipeError:
                    # Handle this error because there's a good chance that godot has crashed.
                    logger.info("caught broken pipe when restarting godot")
            if not self.process.is_closed:
                self.process.close(kill=True, raise_logged_errors=False)
            self._bridge.after_close()
        self.process = InteractiveGodotProcess(self.config, gpu_id=self.gpu_id)
        self._bridge = GodotEnvBridge[ActionType].build_by_starting_process(
            self.process, (self.config.recording_options.resolution_x, self.config.recording_options.resolution_y)
        )

    def _restart_process(self, rebuild_observation_context: bool = False) -> None:
        if self.is_running:
            self.close()
        self.world_generator = self._create_world_generator()
        self.process = InteractiveGodotProcess(self.config, gpu_id=self.gpu_id)
        self._bridge = GodotEnvBridge[ActionType].build_by_starting_process(
            self.process, (self.config.recording_options.resolution_x, self.config.recording_options.resolution_y)
        )
        if rebuild_observation_context:
            self.observation_context = GodotObservationContext(
                observation_type=self.observation_context.observation_type,
                is_space_flattened=self.observation_context.is_space_flattened,
                available_features=self._bridge.query_available_features(),
            )

    @property
    def flattened_observation_keys(self) -> List[str]:
        return self.observation_context.flattened_observation_keys

    @property
    def observation_space(self):
        return self.observation_context.observation_space

    @property
    def is_running(self):
        return self.process.is_running and self._bridge.is_open

    def seed(self, seed: Optional[int] = None):
        if seed is None:
            seed = _randint_of_size(np.int64)
        return self.seed_nicely(seed)

    def seed_nicely(self, episode_seed: int):
        self._latest_episode_seed = episode_seed
        return self._bridge.seed(episode_seed)

    def step(self, action: Dict[str, np.ndarray]):
        observation, goal_progress = self.act(self.action_type.from_input(action))
        lame_observation = self.observation_context.lamify(observation)
        return lame_observation, goal_progress.reward, goal_progress.is_done, goal_progress.log

    def act(self, action: ActionType) -> Tuple[ObservationType, GoalProgressResult]:
        """Same as `step` with observations in the ObservationType format."""
        assert isinstance(action, self.action_type), f"Must pass `{self.action_type}` objects to step"
        assert self.is_reset_called_already, "Must call reset() before calling step() or act()"
        # TODO: probably put this back, make sufficiently precise and condition on is_action_shape_checked
        # assert self.action_space.contains(attr.asdict(action)), f"Invalid action: {action}"

        try:
            feature_data = self._bridge.act(action)
        except GodotError as e:
            # Godot error prevented reading feature data...
            self._check_for_errors_and_collect_artifacts()
            raise
        observation = self._read_observation_reply(feature_data)

        goal_progress = self.goal_evaluator.calculate_goal_progress(observation)
        return observation, goal_progress

    def debug_act(self, action: DebugCameraAction) -> ObservationType:
        assert isinstance(action, DebugCameraAction), f"Must pass `DebugCameraAction` objects to debug_act"
        feature_data = self._bridge.debug_act(action)
        return self._read_observation_reply(feature_data)

    def save_snapshot(self) -> Path:
        assert self.is_reset_called_already, "Cannot take a snapshot without starting a world (via reset)"
        return Path(self._bridge.save_snapshot())

    def load_snapshot(self, snapshot_path: Path) -> Tuple[ObservationType, GoalProgressResult]:
        assert (
            not snapshot_path.is_file()
        ), f"snapshot_path {snapshot_path} should be the path to the directory returned from save_snapshot."
        assert snapshot_path.exists(), f"cannot load snapshot_path {snapshot_path} because it doesn't exist."
        self.is_reset_called_already = True
        feature_data = self._bridge.load_snapshot(snapshot_path.as_posix())

        observation = self._read_observation_reply(feature_data)

        goal_progress = self.goal_evaluator.calculate_goal_progress(observation)
        return observation, goal_progress

    def _read_observation_reply(self, feature_data: FeatureDataDict) -> ObservationType:
        self._latest_screen = feature_data["rgb"] if "rgb" in self.observation_context.selected_features else None

        if self.is_error_log_checked_after_each_step:
            self._check_for_errors_and_collect_artifacts()

        return self.observation_context.make_observation(feature_data)

    def render(self, mode: str = "rgb_array") -> npt.NDArray:  # type: ignore[override]
        assert mode == "rgb_array", "only rgb_array rendering is currently supported"
        if self._latest_screen is None:
            self._latest_screen = self._bridge.render()
            assert type(self._latest_screen) == npt.NDArray
        return self._latest_screen

    def reset(self):
        observation = self.reset_nicely()
        lame_observation = self.observation_context.lamify(observation)
        return lame_observation

    def reset_nicely(self, *, world_id: Optional[int] = None, episode_seed: Optional[int] = None) -> ObservationType:
        self._latest_screen = None
        self.is_reset_called_already = True

        if episode_seed is None:
            episode_seed = _DEFAULT_EPISODE_SEED
        self._latest_episode_seed = episode_seed

        # calls generate batch and creates a level (from somewhere?)
        world_params = self._get_world_params_by_id(world_id)

        # save in case we have an error
        self.recent_worlds.append(world_params)
        if len(self.recent_worlds) > 5:
            self.recent_worlds.popleft()

        self._check_for_errors_and_collect_artifacts()

        initial_state_features = self._bridge.reset(
            self.action_type.get_null_action(), episode_seed, world_params.main_scene_path, STARTING_HIT_POINTS
        )
        observation = self._read_observation_reply(initial_state_features)

        self.goal_evaluator.reset(observation, world_params)
        return observation

    # note: this is for human playback
    def reset_nicely_with_specific_world(
        self,
        *,
        episode_seed: int,
        world_path: Optional[Union[str, Path]] = None,
        world_params: Optional[GeneratedWorldParamsType] = None,
        starting_hit_points: float = 1.0,
    ) -> ObservationType:
        """Specify the world with `world_params` if available (necessary for the default goal evaluator to work),
        and `world_path` if you're loading a custom non-autogenerated level.
        """
        if not world_path:
            assert world_params is not None
            world_path = world_params.output + "/main.tscn"
        world_path = Path(world_path)
        assert world_path.name.endswith(".tscn"), f"{world_path} must be a valid godot scene generated by avalon."
        assert world_path.exists(), f"No such world {world_path}"

        self._latest_screen = None
        self.is_reset_called_already = True
        self._latest_episode_seed = episode_seed

        initial_state_features = self._bridge.reset(
            self.action_type.get_null_action(), episode_seed, str(world_path), starting_hit_points
        )
        observation = self._read_observation_reply(initial_state_features)

        self.process.check_for_errors()

        if world_params:
            self.goal_evaluator.reset(observation, world_params=world_params)
        else:
            self.goal_evaluator.reset(observation, world_params=None)
        return observation

    def close(self) -> None:
        if self._bridge.is_open:
            self._bridge.close()
        if not self.process.is_closed:
            self.process.close(kill=False)
        self._bridge.after_close()
        self.world_generator.close()

    def __del__(self) -> None:
        try:
            if self.is_running:
                self.close()
        except AttributeError:
            # Ignore errors from incomplete init
            pass
        except BrokenPipeError:
            # ignore errors from race conditions if the pipe was broken by a different close
            if self.is_running:
                raise

    def cleanup(self) -> None:
        shutil.rmtree(self.config.dir_root, ignore_errors=True)

    def get_action_log(self) -> "GodotEnvActionLog[ActionType]":
        return GodotEnvActionLog.parse(self.process.action_record_path, self.action_type)

    def _check_for_errors_and_collect_artifacts(self):
        try:
            self.process.check_for_errors()
        except GodotError as ge:
            capture_exception(ge, is_thrown_without_sentry=False)

            tar_path = self.process.save_artifacts(self.recent_worlds)
            logger.warning(f"Godot failed! Saved recent godot levels and logs to {tar_path}")

            if self.is_logging_artifacts_on_error_to_s3:
                logger.warning(f"Uploading artifacts to S3 bucket {self.s3_bucket_name} with key {tar_path}.")
                s3_client = SimpleS3Client(bucket_name=self.s3_bucket_name)
                s3_client.upload_from_file(tar_path, key=Path(tar_path).name)

            if self.is_godot_restarted_on_error:
                logger.warning("Restarting godot!")
                self._restart_godot_quietly()
            else:
                raise
        except ValueError as e:
            if e.args != ("read of closed file",):
                raise e

    def _spawn_fresh_copy_of_env(
        self, run_uuid: str
    ) -> "GodotEnv[ObservationType, ActionType, GeneratedWorldParamsType]":
        "Spawns a new GodotEnv with the same initial arguments as this one."
        return GodotEnv(
            config=self.config,
            observation_type=self.observation_context.observation_type,
            action_type=self.action_type,
            goal_evaluator=self.goal_evaluator,
            gpu_id=self.gpu_id,
            is_logging_artifacts_on_error_to_s3=self.is_logging_artifacts_on_error_to_s3,
            s3_bucket_name=self.s3_bucket_name,
            is_error_log_checked_after_each_step=self.is_error_log_checked_after_each_step,
            is_observation_space_flattened=self.observation_context.is_space_flattened,
            is_godot_restarted_on_error=self.is_godot_restarted_on_error,
            is_dev_flag_added=self.is_dev_flag_added,
            run_uuid=run_uuid,
        )

    def get_replay(self, world_path: Optional[str] = None) -> "GodotEnvReplay":
        from avalon.datagen.godot_env.replay import GodotEnvReplay

        if not self.is_reset_called_already:
            logger.warning(
                f"{self}.get_replay creates a new GodotEnv, but this env is not started. "
                f"Consider using GodotEnvReplay.from_env instead."
            )
        return GodotEnvReplay.from_env(self._spawn_fresh_copy_of_env(self.process.run_uuid), world_path)

    def _get_world_params_by_id(self, world_id: Optional[int]) -> GeneratedWorldParamsType:
        return only(self.world_generator.generate_batch(world_id, 1))


# utility method to get action type given an avalon config. Ideally, this should probably be in a utility class
# but this requires more refactoring due to circular dependency issues


def _randint_of_size(dtype: Type[np.integer]):
    bounds = np.iinfo(dtype)
    return np.random.randint(low=bounds.min, high=bounds.max + 1)



import copy
import json
import os
import uuid
from collections import defaultdict
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import Hashable
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple

import attr
import gym
import numpy as np
import torch
from PIL import Image
from einops import rearrange
from gym import Wrapper
from numpy.typing import NDArray
from torch import Tensor

from avalon.agent.common.params import EnvironmentParams
from avalon.common.errors import SwitchError
from avalon.common.utils import DATA_FOLDER
from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.datagen.godot_base_types import Vector3
from avalon.datagen.godot_env.actions import VRAction
from avalon.datagen.godot_env.goals import AvalonGoalEvaluator
from avalon.datagen.godot_env.goals import GoalProgressResult
from avalon.datagen.godot_env.goals import TrainingAvalonGoalEvaluator
from avalon.datagen.godot_env.godot_env import GodotEnv
from avalon.datagen.godot_env.observations import AvalonObservation
from avalon.datagen.godot_generated_types import AvalonSimSpec
from avalon.datagen.godot_generated_types import RecordingOptionsSpec
from avalon.datagen.godot_generated_types import VRAgentPlayerSpec
from avalon.datagen.godot_utils import S3_AVALON_ERROR_BUCKET
from avalon.datagen.world_creation.constants import TRAIN_TASK_GROUPS
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.constants import AvalonTaskGroup
from avalon.datagen.world_creation.constants import get_all_tasks_for_task_groups
from avalon.datagen.world_creation.constants import int_to_avalon_task
from avalon.datagen.world_creation.entities.spawn_point import PLAYER_SPAWN_POINT
from avalon.datagen.world_creation.world_generator import AvalonWorldGenerator
from avalon.datagen.world_creation.world_generator import BlockingWorldGenerator
from avalon.datagen.world_creation.world_generator import FixedWorldGenerator
from avalon.datagen.world_creation.world_generator import FixedWorldLoader
from avalon.datagen.world_creation.world_generator import GeneratedAvalonWorldParams
from avalon.datagen.world_creation.world_generator import LocalProcessWorldGenerator
from avalon.datagen.world_creation.world_generator import is_fixed_world_generator


def create_base_benchmark_config(
    data_dir: str = str(uuid.uuid4()),
    random_int: int = 0,
    num_frames_alive_after_food_is_gone: int = 20,
    resolution: int = 96,
) -> AvalonSimSpec:
    return AvalonSimSpec(
        episode_seed=0,
        dir_root=os.path.join(DATA_FOLDER, data_dir),
        recording_options=RecordingOptionsSpec(
            user_id=None,
            apk_version=None,
            recorder_host=None,
            recorder_port=None,
            is_teleporter_enabled=False,
            resolution_x=resolution,
            resolution_y=resolution,
            is_recording_human_actions=False,
            is_remote_recording_enabled=False,
            is_adding_debugging_views=False,  # Set to True for debugging videos
            is_debugging_output_requested=False,
        ),
        player=VRAgentPlayerSpec(
            spawn_point_name=PLAYER_SPAWN_POINT,
            max_head_linear_speed=0.3,
            max_head_angular_speed=Vector3(5.0, 10.0, 1.0),
            max_hand_linear_speed=1.0,
            max_hand_angular_speed=15.0,
            jump_height=1.5,
            arm_length=1,
            starting_hit_points=1.0,
            mass=60.0,
            arm_mass_ratio=0.05,
            head_mass_ratio=0.08,
            standup_speed_after_climbing=0.1,
            min_head_position_off_of_floor=0.2,
            push_force_magnitude=5.0,
            throw_force_magnitude=3.0,
            starting_left_hand_position_relative_to_head=Vector3(-0.5, -0.5, -0.5),
            starting_right_hand_position_relative_to_head=Vector3(0.5, -0.5, -0.5),
            minimum_fall_speed=10.0,
            fall_damage_coefficient=0.000026,
            num_frames_alive_after_food_is_gone=num_frames_alive_after_food_is_gone,
            eat_area_radius=0.5,
            is_displaying_debug_meshes=False,  # Set to True for debugging videos
            is_human_playback_enabled=False,
            is_slowed_from_crouching=True,
        ),
    )


LEVEL_OUTPUT_PATH = f"{FILESYSTEM_ROOT}/data/level_gen"
CURRICULUM_BASE_PATH = Path(DATA_FOLDER) / "curriculum"


class TrainingProtocolChoice(Enum):
    MULTI_TASK_ALL = "MULTI_TASK_ALL"
    MULTI_TASK_EASY = "MULTI_TASK_EASY"  # Just MOVE and EAT
    MULTI_TASK_BASIC = "MULTI_TASK_BASIC"  # All non-compositional tasks
    MULTI_TASK_COMPOSITIONAL = "MULTI_TASK_COMPOSITIONAL"
    SINGLE_TASK_EAT = "SINGLE_TASK_EAT"
    SINGLE_TASK_MOVE = "SINGLE_TASK_MOVE"
    SINGLE_TASK_JUMP = "SINGLE_TASK_JUMP"
    SINGLE_TASK_EXPLORE = "SINGLE_TASK_EXPLORE"
    SINGLE_TASK_SCRAMBLE = "SINGLE_TASK_SCRAMBLE"
    SINGLE_TASK_CLIMB = "SINGLE_TASK_CLIMB"
    SINGLE_TASK_DESCEND = "SINGLE_TASK_DESCEND"
    SINGLE_TASK_THROW = "SINGLE_TASK_THROW"
    SINGLE_TASK_AVOID = "SINGLE_TASK_AVOID"
    SINGLE_TASK_HUNT = "SINGLE_TASK_HUNT"
    SINGLE_TASK_FIGHT = "SINGLE_TASK_FIGHT"
    SINGLE_TASK_PUSH = "SINGLE_TASK_PUSH"
    SINGLE_TASK_STACK = "SINGLE_TASK_STACK"
    SINGLE_TASK_BRIDGE = "SINGLE_TASK_BRIDGE"
    SINGLE_TASK_OPEN = "SINGLE_TASK_OPEN"
    SINGLE_TASK_CARRY = "SINGLE_TASK_CARRY"
    SINGLE_TASK_NAVIGATE = "SINGLE_TASK_NAVIGATE"
    SINGLE_TASK_FIND = "SINGLE_TASK_FIND"
    SINGLE_TASK_GATHER = "SINGLE_TASK_GATHER"
    SINGLE_TASK_SURVIVE = "SINGLE_TASK_SURVIVE"


def task_groups_from_training_protocol(
    training_protocol: TrainingProtocolChoice, is_meta_curriculum_used: bool
) -> Tuple[AvalonTaskGroup, ...]:
    if training_protocol == TrainingProtocolChoice.MULTI_TASK_BASIC:
        if is_meta_curriculum_used:
            return TRAIN_TASK_GROUPS[:4]
        else:
            return (AvalonTaskGroup.SIMPLE,)
    elif training_protocol == TrainingProtocolChoice.MULTI_TASK_EASY:
        if is_meta_curriculum_used:
            return TRAIN_TASK_GROUPS[:1]
        else:
            return (AvalonTaskGroup.EASY,)
    elif training_protocol == TrainingProtocolChoice.MULTI_TASK_ALL:
        if is_meta_curriculum_used:
            return TRAIN_TASK_GROUPS
        else:
            return (AvalonTaskGroup.ALL,)
    elif training_protocol == TrainingProtocolChoice.MULTI_TASK_COMPOSITIONAL:
        return (AvalonTaskGroup.COMPOSITIONAL,)
    else:
        assert training_protocol.value.startswith("SINGLE_TASK")
        task_name = training_protocol.value.split("_")[2]
        return (AvalonTaskGroup[task_name],)


@attr.s(auto_attribs=True, frozen=True)
class GodotEnvironmentParams(EnvironmentParams):
    suite: str = "godot"
    # seed: int = 0
    # whether to use per-task curriculum learning. requires a wrapper.
    is_task_curriculum_used: bool = True
    # how much to update the task difficulty per successful or failed episode.
    task_difficulty_update: float = 0.01
    # whether to use a meta-curriculum. requires a wrapper.
    is_meta_curriculum_used: bool = False
    # if using the meta curriculum, the update size
    meta_difficulty_update: float = 0.01
    # initial difficulty value
    initial_difficulty: float = 0
    # can use to give a small per-step penalty for using energy (by moving)
    energy_cost_coefficient: float = 0
    # energy cost for each axis
    head_roll_coefficient: float = 1e-6
    # energy cost for each axis
    head_pitch_coefficient: float = 1e-6
    energy_cost_aggregator: Literal["sum", "max"] = "sum"
    # rgbd image will have resolution (resolution * resolution)
    resolution: int = 96
    greyscale: bool = False
    # dense reward is probably broken; we only use sparse
    is_reward_dense: bool = False
    # the episode will continue for this many steps after eating a food
    num_frames_alive_after_food_is_gone: int = 20
    level_output_base_path: str = f"{FILESYSTEM_ROOT}/data/level_gen"
    # which tasks to train on
    training_protocol: TrainingProtocolChoice = TrainingProtocolChoice.MULTI_TASK_BASIC
    # debug output
    is_debugging_godot: bool = False
    # debug output
    is_video_logged: bool = False
    # debug output
    is_action_logged: bool = False
    # debug output
    is_logging_artifacts_on_error_to_s3: bool = True
    s3_bucket_name: Optional[str] = S3_AVALON_ERROR_BUCKET
    # debug output
    goal_progress_path: Optional[Path] = None

    is_frame_id_transformed_to_frames_remaining: bool = True

    # which for godot to use
    gpu_id: int = 0

    # Evaluation world params
    # fill this in if you want to use fixed worlds for testing
    fixed_worlds_s3_key: Optional[str] = None
    # this will be auto-filled
    fixed_worlds_load_from_path: Optional[Path] = None
    fixed_world_min_difficulty: float = 0
    fixed_world_max_difficulty: float = 1
    val_episodes_per_task: int = 11
    test_episodes_per_task: int = 101
    # difficulty bin size for eval logging histograms
    eval_difficulty_bin_size: float = 0.1

    info_fields: list[str] = [
        "cumulative_episode_return",
        "cumulative_episode_length",
        "task",
        "difficulty",
        "success",
        "score",
        "world_index",
    ]

    @property
    def is_fixed_generator(self) -> bool:
        """Use the `fixed` world genenerator for evaluation."""
        return self.mode != "train"

    @property
    def num_fixed_worlds_per_task(self) -> int:
        """How many evaluation worlds to run per task."""
        if self.mode == "test":
            return self.test_episodes_per_task
        elif self.mode == "val":
            return self.val_episodes_per_task
        else:
            raise SwitchError("mode must be one of {'test', 'val'}")

    @property
    def task_groups(self) -> Tuple[AvalonTaskGroup, ...]:
        return task_groups_from_training_protocol(self.training_protocol, self.is_meta_curriculum_used)

    @property
    def num_tasks(self) -> int:
        return len(get_all_tasks_for_task_groups(self.task_groups))

    @property
    def seed(self) -> int:
        return self.env_index


def write_video_from_np_arrays(video_path: Path, arrays: List[np.ndarray], fps: float = 20) -> None:
    # size = arrays[0].shape[:2]
    # out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*"mp4v"), fps, size, False)
    # for array in arrays:
    #     out.write(array[::-1, :, :3])
    # out.release()
    images = [Image.fromarray(x) for x in arrays]
    with open(video_path, "wb") as f:
        images[0].save(f, save_all=True, append_images=images[1:], optimize=False, duration=100, loop=0)


class AvalonEnv(GodotEnv[AvalonObservation, VRAction, GeneratedAvalonWorldParams]):
    def __init__(self, params: GodotEnvironmentParams) -> None:
        self.params = params

        base_config = create_base_benchmark_config(
            random_int=params.seed,
            num_frames_alive_after_food_is_gone=params.num_frames_alive_after_food_is_gone,
            resolution=params.resolution,
        )
        with base_config.mutable_clone() as config:
            config.recording_options.is_debugging_output_requested = params.is_debugging_godot
            if params.mode == "test" and params.is_video_logged:
                config.player.is_displaying_debug_meshes = True
                config.recording_options.is_adding_debugging_views = True

        self.episode_frames: List[np.ndarray] = []
        if params.mode == "train":
            goal_evaluator: AvalonGoalEvaluator = TrainingAvalonGoalEvaluator(
                energy_cost_coefficient=params.energy_cost_coefficient,
                head_pitch_coefficient=params.head_pitch_coefficient,
                head_roll_coefficient=params.head_roll_coefficient,
            )
        else:
            goal_evaluator = AvalonGoalEvaluator()

        self.current_world_id: int = 0

        self.episode_goal_progress: List[GoalProgressResult] = []
        self.episode_observations: List[AvalonObservation] = []
        self.episode_actions: List[VRAction] = []
        self.eval_world_ids: List[int] = []

        super().__init__(
            config=config,
            observation_type=AvalonObservation,
            action_type=VRAction,
            goal_evaluator=goal_evaluator,
            is_error_log_checked_after_each_step=params.is_debugging_godot,
            is_godot_restarted_on_error=not params.is_debugging_godot,
            gpu_id=params.gpu_id,
            is_logging_artifacts_on_error_to_s3=params.is_logging_artifacts_on_error_to_s3,
            s3_bucket_name=params.s3_bucket_name,
        )
        # We manually indicate to mypy that self.world_generator will be the one we pass above as otherwise it
        # infers the generic type WorldGenerator[GeneratedAvalonWorldParams] and does not locate Avalon-specific methods
        self.world_generator: AvalonWorldGenerator

    def act(self, action: VRAction) -> Tuple[AvalonObservation, GoalProgressResult]:
        observation, goal_progress = super().act(action)
        if self.params.goal_progress_path is not None:
            self.episode_goal_progress.append(goal_progress)
            if self.params.is_video_logged:
                self.episode_observations.append(observation)
            if self.params.is_action_logged:
                self.episode_actions.append(action)
            if goal_progress.is_done:
                assert goal_progress.world_path is not None
                log_goal_progress_path = self.params.goal_progress_path / Path(goal_progress.world_path).relative_to(
                    LEVEL_OUTPUT_PATH
                )
                if not log_goal_progress_path.exists():
                    os.makedirs(log_goal_progress_path)
                    with open(log_goal_progress_path / "goal_progress.json", "w") as f:
                        json.dump([attr.asdict(x) for x in self.episode_goal_progress], f)
                    if self.params.is_video_logged:
                        write_video_from_np_arrays(
                            log_goal_progress_path / "video.png",
                            [
                                np.concatenate([x.rgbd, x.top_down_rgbd, x.isometric_rgbd], axis=1)[::-1, :, :3]
                                for x in self.episode_observations
                            ],
                        )
                        # write_video_from_np_arrays(
                        #     log_goal_progress_path / "top_down_video.png",
                        #     [x.top_down_rgbd[::-1, :, :3] for x in self.episode_observations],
                        # )
                        # write_video_from_np_arrays(
                        #     log_goal_progress_path / "isometric_video.png",
                        #     [x.isometric_rgbd[::-1, :, :3] for x in self.episode_observations],
                        # )
                    if self.params.is_action_logged:
                        with open(log_goal_progress_path / "actions.json", "w") as f:
                            json.dump([attr.asdict(x) for x in self.episode_actions], f)
                self.episode_goal_progress = []
                self.episode_observations = []
                self.episode_actions = []

        return observation, goal_progress

    def update_lame_observation(self, lame_observation: Dict[str, Any]) -> None:
        if self.params.is_frame_id_transformed_to_frames_remaining and "frame_id" in lame_observation:
            assert isinstance(self.goal_evaluator, AvalonGoalEvaluator)
            lame_observation["frame_id"] = self.goal_evaluator.frame_limit - lame_observation["frame_id"] - 1

    def step(self, action: Dict[str, np.ndarray]) -> tuple[AvalonObservation, float, bool, dict]:
        observation, goal_progress = self.act(self.action_type.from_input(action))
        lame_observation = self.observation_context.lamify(observation)
        self.update_lame_observation(lame_observation)
        return lame_observation, goal_progress.reward, goal_progress.is_done, goal_progress.log

    def reset(self):
        world_id = None
        if self.params.is_fixed_generator:
            if len(self.eval_world_ids) == 0:
                assert is_fixed_world_generator(self.world_generator)
                self.eval_world_ids = sorted(self.world_generator.worlds.keys())

            world_id = self.eval_world_ids[self.current_world_id]
            self.current_world_id = (self.current_world_id + 1) % len(self.eval_world_ids)

        observation = self.reset_nicely(world_id=world_id)
        if self.params.is_video_logged:
            self.episode_observations.append(observation)
        lame_observation = self.observation_context.lamify(observation)
        self.update_lame_observation(lame_observation)
        return lame_observation

    def _create_world_generator(self) -> AvalonWorldGenerator:
        base_path = Path(self.params.level_output_base_path)
        if self.params.is_fixed_generator:
            if self.params.fixed_worlds_load_from_path:
                return FixedWorldLoader(
                    base_path=base_path,
                    generator_index=self.params.env_index,
                    num_generators=self.params.env_count,
                    generated_worlds_path=self.params.fixed_worlds_load_from_path,
                )
            difficulties = tuple(
                np.linspace(
                    self.params.fixed_world_min_difficulty,
                    self.params.fixed_world_max_difficulty,
                    self.params.num_fixed_worlds_per_task,
                )
            )
            return FixedWorldGenerator(
                base_path=base_path,
                seed=self.params.seed,
                difficulties=difficulties,
                task_groups=self.params.task_groups,
                generator_index=self.params.env_index,
                num_generators=self.params.env_count,
            )
        return LocalProcessWorldGenerator(
            base_path=base_path,
            seed=self.params.seed,
            min_difficulty=0,
            start_difficulty=self.params.initial_difficulty,
            task_groups=self.params.task_groups,
            is_task_curriculum_used=self.params.is_task_curriculum_used,
        )

    @property
    def curriculum_base_path(self) -> Path:
        # TODO make this not collide
        return CURRICULUM_BASE_PATH

    @property
    def curriculum_save_path(self) -> Path:
        if not self.curriculum_base_path.exists():
            os.makedirs(self.curriculum_base_path, exist_ok=True)
        return self.curriculum_base_path / f"{self.params.env_index}.pkl"

    def __getstate__(self) -> Dict[str, Any]:
        """Make a pickle state for garage to serialize aggressively, not actually the same state, in particular will
        start from the next video

        Returns:
            dict: The pickled state.

        """
        state = self.__dict__.copy()
        # Don't pickle the process
        state["process"] = None
        state["world_generator"] = None
        state["_bridge"] = None
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """See __getstate__ docstring

        Args:
            state (dict): Unpickled state.

        """

        self.__dict__.update(state)
        self._restart_process()
        self.seed_nicely(state["_latest_seed"])

    # task functions from ray.rllib.env.apis.task_settable_env
    def sample_tasks(self, n_tasks: int) -> List[Hashable]:
        if self.params.is_fixed_generator:
            return [0] * n_tasks
        else:
            return self.world_generator.sample_tasks(n_tasks)

    def set_task(self, task: Hashable) -> None:
        if not self.params.is_fixed_generator:
            self.world_generator.set_task(task)
        else:
            assert False

    def set_task_difficulty(self, task: AvalonTask, difficulty: float) -> None:
        assert not self.params.is_fixed_generator
        assert isinstance(self.world_generator, (LocalProcessWorldGenerator, BlockingWorldGenerator))
        self.world_generator.set_task_difficulty(task, difficulty)

    def set_meta_difficulty(self, difficulty: float) -> None:
        assert not self.params.is_fixed_generator
        assert isinstance(self.world_generator, (LocalProcessWorldGenerator, BlockingWorldGenerator))
        self.world_generator.set_meta_difficulty(difficulty)

    def get_task(self) -> Hashable:
        if self.params.is_fixed_generator:
            return 0
        else:
            return self.world_generator.get_task()

    def _get_world_params_by_id(self, world_id: Optional[int]) -> GeneratedAvalonWorldParams:
        if world_id is not None:
            already_generated = self.world_generator.load_already_generated_params(world_id)
            if already_generated is not None:
                return already_generated
        return super()._get_world_params_by_id(world_id)


def _transform_observation_rgbd(x: NDArray, greyscale: bool = False) -> NDArray:
    # flip y
    x = x[::-1]
    if greyscale:
        x = x[:, :, 0:1]
    # torch ordering
    x = rearrange(x, "h w c -> c h w")
    # Copy here because torch doesn't like strided arrays
    return x.copy()


def _transform_observation_rgbd_post(x: Tensor) -> Tensor:
    assert x.dtype in (np.uint8, torch.uint8)
    return x / 255.0 - 0.5


def _transform_observation_scale_bias(x: NDArray, scale: float = 1.0, bias: float = 0.0) -> NDArray:
    return x * scale + bias


def _transform_observation_clamp(x: NDArray, min: Optional[float] = None, max: Optional[float] = None) -> NDArray:
    return np.clip(x, a_min=min, a_max=max)  # type: ignore


class GodotObsTransformWrapper(gym.ObservationWrapper):
    def __init__(self, env: GodotEnv, greyscale: bool = False) -> None:
        super().__init__(env)
        self.transforms = {
            "rgbd": partial(_transform_observation_rgbd, greyscale=greyscale),
            "physical_body_delta_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_body_delta_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_head_delta_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_left_hand_delta_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_right_hand_delta_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_head_delta_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_left_hand_delta_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_right_hand_delta_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_head_relative_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_left_hand_relative_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_right_hand_relative_position": partial(_transform_observation_scale_bias, scale=0.1),
            "physical_head_relative_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_left_hand_relative_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "physical_right_hand_relative_rotation": partial(_transform_observation_scale_bias, scale=0.005),
            "left_hand_thing_colliding_with_hand": partial(_transform_observation_clamp, min=0, max=1),
            "left_hand_held_thing": partial(_transform_observation_clamp, min=0, max=1),
            "right_hand_thing_colliding_with_hand": partial(_transform_observation_clamp, min=0, max=1),
            "right_hand_held_thing": partial(_transform_observation_clamp, min=0, max=1),
            "frame_id": partial(_transform_observation_scale_bias, scale=0.0001),
        }
        assert isinstance(env.observation_space, gym.spaces.Dict)
        self.observation_space = gym.spaces.Dict()
        if "rgbd" in env.observation_space.spaces:
            shape = env.observation_space.spaces["rgbd"].shape
            shape = (1 if greyscale else shape[2], shape[0], shape[1])
            self.observation_space.spaces["rgbd"] = gym.spaces.Box(low=0, high=255, shape=shape, dtype=np.uint8)

        # really these are vector keys
        self.scalar_keys = {
            k: v.shape[0]
            for k, v in env.observation_space.spaces.items()
            if isinstance(v, gym.spaces.Box) and len(v.shape) == 1
        }
        self.vec_len = sum(self.scalar_keys.values())
        assert len(env.observation_space.spaces) == len(self.observation_space.spaces) + len(self.scalar_keys)
        self.observation_space["scalars"] = gym.spaces.Box(low=-1, high=1, shape=(self.vec_len,), dtype=np.float32)

    def observation(self, observation: Dict[str, Any]):
        transformed = {k: self.transforms[k](v) if k in self.transforms else v for k, v in observation.items()}
        out = {}
        if "rgbd" in transformed:
            out["rgbd"] = transformed["rgbd"]
        scalar_outputs = np.zeros((self.vec_len,), dtype=np.float32)
        i = 0
        for k, v in transformed.items():
            if k == "rgbd":
                continue
            scalar_outputs[i : i + len(v)] = v
            i += len(v)
        # Enforce the bounds of the scalar space
        scalar_outputs = np.clip(
            scalar_outputs, self.observation_space["scalars"].low, self.observation_space["scalars"].high  # type: ignore
        )
        out["scalars"] = scalar_outputs
        return out


def squash_real(x: np.ndarray) -> np.ndarray:
    x = np.clip(x, -4, 4)
    return np.tanh(x)


class ScaleAndSquashAction(gym.ActionWrapper):
    """Constrain to within [-1, 1] with a tanh and then scale the actions."""

    def __init__(self, env: gym.Env, key: str = "real", scale: float = 1) -> None:
        super().__init__(env)
        self.key = key
        self.scale = scale

    def action(self, action: Dict[str, np.ndarray]):
        action_dict = copy.deepcopy(action)
        action_dict[self.key] = squash_real(action_dict[self.key] * self.scale)
        return action_dict

    def reverse_action(self, action: Dict[str, np.ndarray]):
        raise NotImplementedError


# Note: the rate of update of difficulty (per global env step) depends on the number of workers.
class CurriculumWrapper(Wrapper):
    def __init__(self, env: AvalonEnv, task_difficulty_update: float, meta_difficulty_update: float) -> None:
        super().__init__(env)
        self._env = env
        self.difficulties: DefaultDict[AvalonTask, float] = defaultdict(float)
        self.task_difficulty_update = task_difficulty_update
        self.meta_difficulty_update = meta_difficulty_update
        self.meta_difficulty = 0.0

    def step(self, action: Dict[str, torch.Tensor]):
        observation, reward, done, info = self.env.step(action)  # type: ignore[misc]
        if done:
            task = AvalonTask[int_to_avalon_task[int(info["task"])]]
            update_step = self.task_difficulty_update  # * np.random.uniform()
            if info["success"] == 1:
                self.difficulties[task] += update_step
                self.meta_difficulty += self.meta_difficulty_update
            elif info["success"] == 0:
                self.difficulties[task] -= update_step
                self.meta_difficulty -= self.meta_difficulty_update
            else:
                assert False, info["success"]
            self.difficulties[task] = max(min(self.difficulties[task], 1.0), 0.0)
            self.meta_difficulty = max(min(self.meta_difficulty, 1.0), 0.0)
            self._env.set_task_difficulty(task, self.difficulties[task])
            self._env.set_meta_difficulty(self.meta_difficulty)
        return observation, reward, done, info



import signal
import time
import uuid
from abc import ABC
from abc import abstractmethod
from ctypes import c_int
from functools import partial
from pathlib import Path
from typing import Generic
from typing import Iterator
from typing import List
from typing import Optional
from typing import Protocol

import attr
import torch
import wandb
from loguru import logger
from torch.utils.data import DataLoader
from tree import map_structure

from avalon.agent.common import wandb_lib
from avalon.agent.common.action_model import visualize_actions
from avalon.agent.common.dataloader import ReplayDataset
from avalon.agent.common.dataloader import worker_init_fn
from avalon.agent.common.envs import build_env
from avalon.agent.common.get_algorithm_cls import get_algorithm_cls
from avalon.agent.common.storage import FragmentStorage
from avalon.agent.common.storage import ModelStorage
from avalon.agent.common.storage import TrajectoryStorage
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.types import ParamsType
from avalon.agent.common.types import StepData
from avalon.agent.common.util import get_checkpoint_file
from avalon.agent.common.util import pack_1d_list
from avalon.agent.common.util import postprocess_uint8_to_float
from avalon.agent.common.worker import AsyncRolloutManager
from avalon.agent.common.worker import RolloutManager
from avalon.agent.dreamer.params import OffPolicyParams
from avalon.agent.godot.godot_eval import log_rollout_stats_packed
from avalon.agent.ppo.params import OnPolicyParams
from avalon.agent.ppo.params import PPOParams
from avalon.agent.ppo.ppo_types import PPOStepData
from avalon.contrib.utils import make_deterministic

# Just a temporary reminder to not use <1.11
assert int(torch.__version__.split(".")[1]) >= 11


class Cleanable(Protocol):
    def shutdown(self) -> None:
        pass


class Trainer(ABC, Generic[ParamsType]):
    def __init__(self, params: ParamsType) -> None:
        # TODO: fix this properly
        import openturns

        openturns.Tensor
        # Patch the issue where openturns resets the signal handler
        signal.signal(signal.SIGINT, signal.default_int_handler)

        # Use an explict context to fix this bug: https://github.com/pytorch/pytorch/issues/3492
        # Spawn is less fiddly, but fork launches faster
        self.multiprocessing_context = torch.multiprocessing.get_context(params.multiprocessing_mode)
        # Fixing the bug where torch gets really slow on small fast ops on CPU
        # https://github.com/pytorch/pytorch/issues/80777
        torch.set_num_threads(1)

        # allow tf32 (also need to call this in any subprocesses that may need it, eg inference)
        # leaving this off for now since it gave issues in PPO
        # torch.backends.cuda.matmul.allow_tf32 = True
        # torch.backends.cudnn.allow_tf32 = True

        self.params = params
        self.to_cleanup: List[Cleanable] = []

        self.start: Optional[float] = None
        self.i = 0
        self.env_step = 0

        if isinstance(params, PPOParams):
            self.step_data_type = PPOStepData
        else:
            self.step_data_type = StepData

        self.get_spaces()
        self.train_storage = self.create_train_storage()

        if params.deterministic:
            make_deterministic(seed=0)

        self.algorithm = self.create_model()
        self.train_rollout_manager = self.create_rollout_manager()
        self.train_dataloader = self.create_dataloader()
        self.wandb_run = self.wandb_init()
        # wandb.watch(self.algorithm, log=None, log_freq=self.params.log_freq_hist, log_graph=True)

    def get_spaces(self) -> None:
        dummy_env = build_env(self.params.env_params)
        self.params = attr.evolve(
            self.params, observation_space=dummy_env.observation_space, action_space=dummy_env.action_space
        )
        dummy_env.close()

    def create_train_storage(self) -> Optional[TrajectoryStorage]:
        return None

    def wandb_init(self):
        # This needs to happen after all other processes launch
        run_name = (
            f"{self.params.env_params.suite}:{self.params.env_params.task}"
            if not self.params.name
            else self.params.name
        )
        run = wandb.init(
            name=run_name,
            project=self.params.project,
            config=attr.asdict(self.params),
            tags=self.params.tags,
            group=self.params.group,
            mode=self.params.wandb_mode,
        )
        wandb_lib.SCALAR_FREQ = self.params.log_freq_scalar
        wandb_lib.HIST_FREQ = self.params.log_freq_hist
        wandb_lib.MEDIA_FREQ = self.params.log_freq_media
        return run

    @abstractmethod
    def create_rollout_manager(self):
        raise NotImplementedError

    def create_model(self) -> Algorithm:
        algorithm_cls = get_algorithm_cls(self.params)
        assert self.params.observation_space is not None
        assert self.params.action_space is not None
        algorithm = algorithm_cls(self.params, self.params.observation_space, self.params.action_space)

        if self.params.resume_from:
            checkpoint_path = get_checkpoint_file(self.params.resume_from)
            algorithm.load_state_dict(torch.load(checkpoint_path, map_location=self.params.train_device))
            logger.info("RESUMED MODEL FROM CHECKPOINT")

        algorithm = algorithm.to(self.params.train_device)

        logger.info(f"model has {sum(p.numel() for p in algorithm.parameters() if p.requires_grad)} params")
        return algorithm

    @abstractmethod
    def create_dataloader(self) -> Iterator:
        raise NotImplementedError

    def train(self) -> None:
        if not self.start:
            self.start = time.time()
        while True:
            self.train_step()

            wandb_lib.log_scalar("env_step", self.env_step, self.i)
            if self.env_step >= self.params.total_env_steps:
                break

        self.checkpoint(filename="final.pt")

    @property
    def frames_per_batch(self):
        return self.params.batch_size

    def train_step(self) -> None:
        rollouts: BatchSequenceData = next(self.train_dataloader)
        start = time.time()
        start_i = self.i
        rollouts = map_structure(lambda x: x.to(self.params.train_device, non_blocking=True), rollouts)
        rollouts = attr.evolve(
            rollouts,
            observation=postprocess_uint8_to_float(rollouts.observation, center=self.params.center_observations),
        )

        self.i = self.algorithm.train_step(rollouts, self.i)

        wandb_lib.log_scalar(
            "timings/train_fps", self.frames_per_batch * (self.i - start_i) / (time.time() - start), self.i
        )
        wandb_lib.log_scalar("true_env_step", self.env_step * self.params.env_params.action_repeat, self.i)

        if self.i % self.params.checkpoint_every == 0:
            self.checkpoint()

        wandb_lib.log_iteration_time(self.frames_per_batch, self.i)
        wandb_lib.log_scalar("timings/cumulative_env_fps", self.env_step / (time.time() - self.start), self.i)
        wandb_lib.log_scalar(
            "timings/cumulative_train_fps", self.i * self.frames_per_batch / (time.time() - self.start), self.i
        )

        # Visualize the observations
        if self.i % wandb_lib.MEDIA_FREQ == 0:
            # the post-transforms have already been applied.
            for k, v in self.params.observation_space.spaces.items():
                if len(v.shape) == 3:
                    obs_video = rollouts.observation[k][:8]
                    wandb_lib.log_video(f"videos/obs/{k}", obs_video + 0.5, self.i, freq=1)

        # Visualize the actions
        if self.i % wandb_lib.MEDIA_FREQ == 0:
            visualize_actions(self.params.action_space, rollouts.action, prefix="rollout_actions", freq=1)

        # Evaluation
        # val.update(algorithm, i, env_step)
        # self.i += 1

    def checkpoint(self, filename: Optional[str] = None) -> None:
        if not filename:
            filename = f"model_{self.i}.pt"
        model_filename = Path(wandb.run.dir) / filename  # type: ignore
        torch.save(self.algorithm.state_dict(), model_filename)
        wandb.save(str(model_filename), policy="now")  # type: ignore
        wandb_lib.log_scalar("last_checkpoint", self.i, self.i, freq=1)

    def test(self):
        if not self.params.is_train_only:
            raise NotImplementedError

    def shutdown(self, finish_wandb_quietly=False) -> None:
        # The main thread won't join until we close all the processes we have open.
        # TODO: we don't automatically cleanup the off-policy dataloader(s)
        for item in self.to_cleanup:
            logger.info(f"shutting down {item}")
            item.shutdown()
        self.to_cleanup = []
        if finish_wandb_quietly:
            logger.info("finishing wandb")
            wandb.finish(quiet=True)
        logger.info("trainer shutdown is finished")


class OffPolicyTrainer(Trainer[OffPolicyParams]):
    def __init__(self, params: OffPolicyParams) -> None:
        self.train_rollout_dir = str(Path(params.data_dir) / "train" / str(uuid.uuid4()))

        # Necessary to fix some bug: https://github.com/wandb/client/issues/1994
        # Also necessary to get wandb to shut down properly?
        wandb.require(experiment="service")
        wandb.setup()

        self.train_rollout_managers: list[AsyncRolloutManager] = []
        super().__init__(params)
        ModelStorage.clean()

    def log_rollout_stats(self) -> None:
        for manager in self.train_rollout_managers:
            wandb_lib.log_from_queue(manager.wandb_queue, prefix=f"rollout_manager_{manager.rollout_manager_id}/")

    def start_train_rollouts(self) -> None:
        [manager.start() for manager in self.train_rollout_managers]

    def shutdown_train_rollouts(self) -> None:
        [manager.shutdown() for manager in self.train_rollout_managers]
        self.train_rollout_managers = []

    def create_rollout_manager(self) -> None:
        self.env_step_counters = []
        for i in range(self.params.worker_managers):
            # we don't need a lock here because there's only one writer and precise read/write ordering doesn't matter
            env_step_counter = self.multiprocessing_context.Value(c_int, lock=False)
            train_game_player = AsyncRolloutManager(
                params=self.params,
                obs_space=self.params.observation_space,
                action_space=self.params.action_space,
                rollout_manager_id=i,
                env_step_counter=env_step_counter,
                multiprocessing_context=self.multiprocessing_context,
                train_rollout_dir=self.train_rollout_dir,
            )
            self.env_step_counters.append(env_step_counter)
            self.to_cleanup.append(train_game_player)
            self.train_rollout_managers.append(train_game_player)

    def create_dataloader(self) -> Iterator[BatchSequenceData]:
        train_dataset = ReplayDataset(self.params, self.train_rollout_dir, update_interval=4000)
        return iter(
            DataLoader(
                train_dataset,
                batch_size=self.params.batch_size,
                shuffle=False,
                num_workers=self.params.num_dataloader_workers,
                drop_last=True,
                pin_memory=True,
                prefetch_factor=2,
                worker_init_fn=worker_init_fn,
                collate_fn=partial(pack_1d_list, out_cls=BatchSequenceData),
            )
        )

    def train_step(self) -> None:
        old_i = self.i
        super().train_step()
        assert self.i == old_i + 1, "Off-policy algorithms must increment i by only 1"
        self.env_step = sum([x.value for x in self.env_step_counters])
        if self.i % self.params.rollout_model_update_freq == 0:
            logger.debug("saving model")
            ModelStorage.push_model(self.algorithm)
        self.log_rollout_stats()

    @property
    def frames_per_batch(self):
        assert self.params.min_fragment_len == self.params.max_fragment_len
        return self.params.batch_size * self.params.min_fragment_len


class OnPolicyTrainer(Trainer[OnPolicyParams]):
    def __init__(
        self, params: OnPolicyParams, rollout_manager_cls=RolloutManager, storage_cls=FragmentStorage
    ) -> None:
        self.rollout_manager_cls = rollout_manager_cls
        self.storage_cls = storage_cls
        super().__init__(params)

    def create_rollout_manager(self):
        rollout_manager = self.rollout_manager_cls(
            params=self.params,
            num_workers=self.params.num_workers,
            is_multiprocessing=self.params.multiprocessing,
            storage=self.train_storage,
            obs_space=self.params.observation_space,
            model=self.algorithm,
            rollout_device=self.params.inference_device,
            multiprocessing_context=self.multiprocessing_context,
        )
        self.to_cleanup.append(rollout_manager)
        return rollout_manager

    def create_train_storage(self) -> TrajectoryStorage:
        return self.storage_cls(self.params, step_data_type=self.step_data_type, num_workers=self.params.num_workers)

    def rollout_step(self) -> None:
        # Run rollouts
        start = time.time()
        self.train_rollout_manager.run_rollout(
            num_steps=self.params.num_steps,
            exploration_mode="explore",
        )
        rollout_fps = (self.params.num_workers * self.params.num_steps) / (time.time() - start)
        wandb_lib.log_scalar("timings/rollout_fps", rollout_fps, self.i)
        self.env_step += self.params.num_workers * self.params.num_steps

        if self.params.env_params.suite == "godot":
            log_rollout_stats_packed(self.train_storage.storage, self.i)
        else:
            # Either use our default worker, or wrap in our (not gym's!) RecordEpisodeStatistics for this to work.
            try:
                if "info" in self.train_storage.storage.keys():
                    rollouts = self.train_storage.storage
                    cumulative_rewards = []
                    episode_lengths = []
                    for worker_id in range(self.params.num_workers):
                        for t in range(self.params.num_steps):
                            if rollouts["done"][worker_id, t]:
                                cumulative_rewards.append(rollouts["info"]["cumulative_episode_return"][worker_id, t])
                                episode_lengths.append(rollouts["info"]["cumulative_episode_length"][worker_id, t])
                    if len(cumulative_rewards) > 0:
                        wandb_lib.log_histogram(f"rollout/cumulative_reward", cumulative_rewards, self.i)
                        wandb_lib.log_histogram(f"rollout/episode_length", episode_lengths, self.i)
                    wandb_lib.log_scalar(f"rollout/num_finished_episodes", len(episode_lengths), self.i)
            except (AttributeError, TypeError):
                # Storage doesn't support this
                pass

    def create_dataloader(self) -> Iterator[BatchSequenceData]:
        def dataloader():
            while True:
                self.train_storage.reset()
                self.rollout_step()
                rollouts: BatchSequenceData = self.train_storage.to_packed()
                # go ahead and send to cuda now, will make the next step faster
                rollouts = map_structure(lambda x: x.to(self.params.train_device, non_blocking=True), rollouts)
                yield rollouts

        return iter(dataloader())



from typing import Union
from typing import cast

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DListNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.height_map import HeightMap


def normalized(v: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    norm = np.linalg.norm(v)
    if norm == 0:
        return v
    return v / norm


def facing_2d(from_point: Point3DNP, to_point: Point3DNP) -> Rotation:
    dir = to_point - from_point
    dir[1] = 0
    dir = -normalized(dir)
    yaw = np.arctan2(dir[0], dir[2])
    return Rotation.from_euler("y", yaw)


def get_random_ground_points(
    rand: np.random.Generator,
    center: Point3DNP,
    count: int,
    map: HeightMap,
    min_radius: float,
    max_radius: float,
    offset: Union[float, np.ndarray],
    island_mask: MapBoolNP,
) -> Point3DListNP:
    assert count > 0
    acceptable_points = None
    scale = 1.0
    for i in range(10):
        radius = rand.uniform(min_radius, max_radius, size=count * 2 + 5)
        angle = rand.uniform(0, np.pi * 2, size=count * 2 + 5)
        x = np.cos(angle) * radius + center[0]
        y = np.sin(angle) * radius + center[2]
        points = np.stack([x, y], axis=1)
        points = map.restrict_points_to_region(points)
        idx_x, idx_y = map.points_to_indices(points)
        points = points[island_mask[idx_x, idx_y]]
        heights = map.get_heights(points)
        points_3d = np.stack([points[:, 0], heights, points[:, 1]], axis=1)
        if acceptable_points is None:
            acceptable_points = points_3d
        else:
            acceptable_points = np.concatenate([acceptable_points, points_3d], axis=0)
        if acceptable_points.shape[0] >= count:
            acceptable_points = acceptable_points[:count, :]
            acceptable_points[:, 1] += scale * offset
            return cast(Point3DListNP, acceptable_points)
    raise ImpossibleWorldError("Unable to create enough random points. Likely in the water.")



from typing import Tuple
from typing import TypeVar

import attr
import numpy as np
from godot_parser import Node as GDNode

from avalon.common.utils import to_immutable_array
from avalon.contrib.serialization import Serializable
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Entity(Serializable):
    entity_id: int
    is_dynamic: bool
    position: np.ndarray = attr.ib(
        converter=to_immutable_array, eq=attr.cmp_using(eq=np.array_equal)  # type: ignore[attr-defined]
    )

    @property
    def point2d(self) -> Tuple[float, float]:
        return (self.position[0], self.position[2])

    def get_node(self, scene: GodotScene) -> GDNode:
        raise NotImplementedError()


EntityType = TypeVar("EntityType", bound=Entity)



import os
import pickle
import warnings
from functools import partial
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import attr
import gym
import numpy as np
import torch
from gym.spaces import Box
from numpy.typing import NDArray
from torch import Tensor
from tree import map_structure

from avalon.agent.common import wandb_lib
from avalon.agent.common.wrappers import OneHotMultiDiscrete
from avalon.common.type_utils import assert_not_none
from avalon.contrib.utils import make_deterministic

numpy_to_torch_dtype = {
    bool: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
}


def get_checkpoint_file(protocol_path: str) -> str:
    """Gets a string path to the checkpoint file.
    - protocol_path:
        a string specifying the possible places to resume from, either local file or wandb run
        valid formats include:
            wandb://{project}/{run}/{file_name}
            file://{absolute_file_path}
        for example:
            wandb://untitled-ai/sf3189ytcfg/checkpoint.pt
            file:///home/user/runs/good_run/checkpoint.pt
    """
    protocol, path = protocol_path.split("://")
    if protocol == "wandb":
        project, run_id, filename = path.split("/")
        return wandb_lib.download_file(run_id, project, filename)
    elif protocol == "file":
        return path
    else:
        raise ValueError(f"protocol {protocol} in {protocol_path} is not currently supported")


PackedType = TypeVar("PackedType")


def pack_1d_list(sequence: List, out_cls: Type[PackedType]) -> PackedType:
    """Pack a list of StepDatas into a BatchData, or a list of SequenceDatas into a BatchSequenceData"""
    out: dict[str, Any] = {}
    for field_obj in attr.fields(type(sequence[0])):
        field = field_obj.name
        if field == "info":
            continue
        sample = getattr(sequence[0], field)
        if isinstance(sample, dict):
            out[field] = {
                k: torch.stack([getattr(transition, field)[k] for transition in sequence]) for k in sample.keys()
            }
        elif np.isscalar(sample):
            out[field] = torch.tensor(np.stack([getattr(transition, field) for transition in sequence]))
        else:
            # tensor
            out[field] = torch.stack([getattr(transition, field) for transition in sequence])
    return out_cls(**out)


def pack_2d_list(batch: List[List], out_cls: Type[PackedType]) -> PackedType:
    """Pack a batch of StepDatas into a BatchSequenceData (or subclass)"""
    out: dict[str, Any] = {}
    for k_obj in attr.fields(type(batch[0][0])):
        k = k_obj.name
        if k == "info":
            continue
        example = batch[0][0].__getattribute__(k)
        if np.isscalar(example):
            out[k] = torch.tensor(
                np.stack([np.stack([getattr(transition, k) for transition in trajectory]) for trajectory in batch])
            )
        elif isinstance(example, Tensor):
            out[k] = torch.stack(
                [torch.stack([getattr(transition, k) for transition in trajectory]) for trajectory in batch]
            )
        elif isinstance(example, dict):
            out2: dict[str, Any] = {}
            for k2 in example.keys():
                out2[k2] = torch.stack(
                    [torch.stack([getattr(transition, k)[k2] for transition in trajectory]) for trajectory in batch]
                )
            out[k] = out2
        else:
            assert False
    return out_cls(**out)


def postprocess_uint8_to_float(
    data: Dict[str, torch.Tensor], center: bool = True, observation_prefix: str = ""
) -> Dict[str, torch.Tensor]:
    """Convert uint8 (0,255) to float (-.5, .5) in a dictionary of rollout data.

    We use this to keep images as uint8 in storage + transfer.
    If observation_prefix is passed, then only keys with that prefix will be looked at for transforming.
    This allows handling a whole batch of rollout data (with actions, values, etc), which is convenient.
    """
    out = {}
    for k, v in data.items():
        if k.startswith(observation_prefix) and v.dtype in (np.uint8, torch.uint8):
            v = v / 255.0
            if center:
                v = v - 0.5
        out[k] = v
    return out


ArrayType = Union[NDArray, torch.Tensor]


def hash_tensor(x: Tensor, data_only: bool = False) -> str:
    """Get a unique hash for a tensor.
    If `include_metadata` is False, the hash is based only on the data itself (and dtype, and shape).
    If `include_metadata` is True, the hash includes all things known to make 2 tensors act differently.
    """
    to_hash = str(pickle.dumps(torch.clone(x.detach().cpu(), memory_format=torch.contiguous_format).numpy()))
    if not data_only:
        to_hash += str(x.stride())  # stride matters for determinism and isn't otherwise checked
        to_hash += str(x.device)
        to_hash += str(x.dtype)
    return str(hash(to_hash))[:8]


def debug_hash_tensor(name: str, x: Tensor, concise: bool = True, print_tensor: bool = False) -> None:
    """Print useful output to help understand why two tensors don't have the same hash."""
    torch.set_printoptions(precision=8)
    data_only = torch.clone(x.cpu(), memory_format=torch.contiguous_format)
    print(
        f"{name}; full hash: {hash_tensor(x)}; data hash: {hash_tensor(data_only, data_only=True)}; shape: {x.shape}; dtype: {x.dtype}; device: {x.device}"
    )
    if not concise:
        print("sum", x.sum())
        print("device", x.device)
        print("stride", x.stride())
        print("max", x.max())
        print("min", x.min())
    if print_tensor:
        print(x)


class HashTensor(torch.nn.Module):
    """A module to inject hashing debug output into a torch.Sequential or similar."""

    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def forward(self, x: Tensor) -> Tensor:
        print(self.name, hash_tensor(x))
        return x


def hash_model(model: torch.nn.Module) -> str:
    return str(hash(tuple(hash_tensor(p.data, data_only=True) for p in model.state_dict().values())))[:8]


def explained_variance(y_pred: ArrayType, y_true: ArrayType) -> float:
    """
    Computes fraction of variance that ypred explains about y.
    Returns 1 - Var[y-ypred] / Var[y]

    interpretation:
        ev=0  =>  might as well have predicted zero
        ev=1  =>  perfect prediction
        ev<0  =>  worse than just predicting zero

    :param y_pred: the prediction
    :param y_true: the expected value
    :return: explained variance of ypred and y
    """
    if isinstance(y_pred, torch.Tensor):
        y_pred = y_pred.flatten()
        y_true = y_true.flatten()
        assert len(y_pred) == len(y_true)
        var_y = torch.var(y_true)  # type: ignore
        return np.nan if var_y == 0 else 1 - torch.var(y_true - y_pred).item() / var_y  # type: ignore
    elif isinstance(y_pred, np.ndarray):
        y_pred = y_pred.flatten()
        y_true = y_true.flatten()
        assert len(y_pred) == len(y_true)
        var_y = np.var(y_true)
        return np.nan if var_y == 0 else 1 - np.var(y_true - y_pred).item() / var_y  # type: ignore
    else:
        import tensorflow as tf

        if tf.is_tensor(y_pred):
            y_pred = tf.reshape(y_pred, [-1])
            y_true = tf.reshape(y_true, [-1])
            assert len(y_pred) == len(y_true)
            var_y = tf.math.reduce_variance(y_true)
            return np.nan if var_y == 0 else 1 - tf.math.reduce_variance(y_true - y_pred).numpy() / var_y
        else:
            raise ValueError


def create_action_storage(
    action_space: gym.spaces.Dict, batch_shape: tuple[int, ...], use_shared_memory: bool = False
) -> dict[str, Tensor]:
    action: dict[str, Tensor] = {}
    for k, v in action_space.items():
        if isinstance(v, OneHotMultiDiscrete):
            action[k] = torch.zeros(
                (*batch_shape, *assert_not_none(v.shape), v.max_categories),
                dtype=numpy_to_torch_dtype[assert_not_none(v.dtype).type],
            )
            if use_shared_memory:
                action[k] = action[k].share_memory_()
        elif isinstance(v, Box):
            action[k] = torch.zeros(
                (*batch_shape, *assert_not_none(v.shape)), dtype=numpy_to_torch_dtype[assert_not_none(v.dtype).type]
            )
            if use_shared_memory:
                action[k] = action[k].share_memory_()
        else:
            raise NotImplementedError
    return action


def create_observation_storage(
    observation_space: gym.spaces.Dict, batch_shape: tuple[int, ...], use_shared_memory: bool = False
) -> dict[str, Tensor]:
    observation: dict[str, Tensor] = {}
    for k, v in observation_space.items():
        observation[k] = torch.zeros(
            (*batch_shape, *assert_not_none(v.shape)),
            dtype=numpy_to_torch_dtype[v.dtype.type],
        )
        if use_shared_memory:
            observation[k] = observation[k].share_memory_()
    return observation


def copy(target: ArrayType, source: ArrayType) -> None:
    target[:] = source  # type: ignore


def masked_copy(target: ArrayType, source: ArrayType, mask: ArrayType) -> None:
    target[mask] = source[mask]  # type: ignore


def masked_copy_structure(target: ArrayType, source: ArrayType, mask: Optional[ArrayType]) -> None:
    """copy each atom of `source` into target, where `mask` is true"""
    if mask is not None:
        map_structure(partial(masked_copy, mask=mask), target, source)
    else:
        map_structure(copy, target, source)


def get_avalon_model_seed() -> Optional[int]:
    seed = os.getenv("AVALON_MODEL_SEED")
    return int(seed) if seed is not None else None


def seed_and_run_deterministically_if_enabled() -> None:
    if (seed := get_avalon_model_seed()) is not None:
        make_deterministic(seed)


def setup_new_process() -> None:
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    torch.set_num_threads(1)



import json
import os
import platform
import sys
from pathlib import Path
from typing import Dict
from typing import Optional

from loguru import logger
from tqdm import tqdm

from avalon.common.error_utils import capture_exception
from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.contrib.utils import is_notebook

__all__ = [
    "logger",
    "configure_local_logger",
    "configure_remote_logger",
    "get_log_path",
    "get_user_name",
    "get_experiment_name",
    "log_to_sentry",
    "log_to_sentry_and_reraise",
]


def get_user_name() -> str:
    return os.getenv("EXPERIMENT_USERNAME", "default_user")


def get_experiment_name() -> str:
    return os.getenv("EXPERIMENT_NAME", "default_experiment")


def get_log_path() -> Path:
    if platform.system().lower() == "darwin":
        log_folder = Path("/tmp/logs")
    else:
        log_folder = Path(FILESYSTEM_ROOT) / "logs"
    log_folder.mkdir(parents=True, exist_ok=True)
    return log_folder


def log_to_sentry(exception: BaseException, extra: Optional[Dict[str, str]] = None):
    if is_notebook():
        raise exception

    if extra is None:
        extra = {}

    user = f"{get_user_name()}-sentry"
    experiment = get_experiment_name()
    # passing in extra fingerprint of user so that issues on sentry are unique per user
    capture_exception(exception, extra=dict(username=user, experiment=experiment, **extra), extra_fingerprint=user)


def log_to_sentry_and_reraise(exception: BaseException):
    log_to_sentry(exception)
    raise exception


tqdm.get_lock()
# disable monitoring thread because it is not totally safe, caused some issues with multithreading.
tqdm.monitor_interval = 0


def configure_local_logger(level: str = "DEBUG", format: Optional[str] = None) -> None:
    try:
        from computronium.common.log_utils import configure_local_logger as inner_configure_local_logger
    except ImportError:
        logger.remove()
        logger.add(sys.stdout, level=level, format="{message}")
        os.environ["_LOG_CONFIG"] = json.dumps(dict(level=level, format=format))
    else:
        if format is None:
            inner_configure_local_logger(level)
        else:
            inner_configure_local_logger(level, format)


def configure_remote_logger(level: str = "DEBUG", format: Optional[str] = None) -> None:
    try:
        from computronium.common.log_utils import configure_remote_logger as inner_configure_remote_logger
    except ImportError:
        logger.remove()
        logger.add(sys.stdout, level=level, format="{message}")
        os.environ["_LOG_CONFIG"] = json.dumps(dict(level=level, format=format))
    else:
        if format is None:
            inner_configure_remote_logger(level)
        else:
            inner_configure_remote_logger(level, format)


def configure_parent_logging() -> None:
    try:
        from computronium.common.log_utils import configure_parent_logging as inner_configure_parent_logging
    except ImportError:
        if "_LOG_CONFIG" not in os.environ:
            raise Exception(
                "No parent process logging configuration specified. You are likely attempting to call configure_parent_logging without having already called configure_remote_logger or configure_local_logger in a parent process. If this is not a spawned subprocess, you should call one of those two functions rather than calling this function."
            )
        log_config = json.loads(os.environ["_LOG_CONFIG"])
        configure_local_logger(**log_config)
    else:
        inner_configure_parent_logging()
